#!/usr/bin/perl -w
# -*-Perl-*-
# Last changed Time-stamp: <2010-10-28 14:15:39 jana>
#
use strict;
use Getopt::Long;
use Data::Dumper;

my ($in_file, $help, $rmGaps);
$rmGaps = 0;

my $usage= << "JUS";

  USAGE:    perl $0 -i FASTAfile
    
  options:  -i file:  FASTA formatted file of aligned sequences
            -rmGaps   remove gap columns
    
  note:     bla..

  results:  Program $0 bla..
    
  example: perl $0 -i 18S.structures.fasta
    
JUS

print $usage unless GetOptions("h|?" => \$help,
                          "i=s" => \$in_file,
			       "rmGaps" => \$rmGaps
                          );
if($help || !($in_file)) { print $usage; exit(1); }

open(F, $in_file) or die $!; my @fasta = <F>; close(F);

my ($seq, %freq, $i);

my (%seqs, $id);
for($i=0; $i<scalar(@fasta); $i++) {
    if($fasta[$i] =~ /^>(.+)$/) {
	$id = $1;
	$seqs{$id} = "";
    } else {
	chomp($fasta[$i]);
	$seqs{$id} .= uc $fasta[$i];
    }
}

my $l = length $seqs{$id};
my $n = scalar keys %seqs;

for($i=0; $i<$l; $i++) {
    $freq{$i}{A} = 0;
    $freq{$i}{C} = 0;
    $freq{$i}{G} = 0;
    $freq{$i}{U} = 0;
    $freq{$i}{X} = 0;
    $freq{$i}{'-'} = 0;
}

foreach my $k (keys %seqs) {
    $seq = $seqs{$k};	
    my @pos = split //,$seq;
    for(my $i=0; $i<scalar(@pos); $i++) {
	if($pos[$i] eq '.') {
	    $freq{$i}{'-'}++;
	} elsif($pos[$i] eq 'N') {
	    $freq{$i}{A} += 1/4;
	    $freq{$i}{C} += 1/4;
	    $freq{$i}{G} += 1/4;
	    $freq{$i}{U} += 1/4;
	} elsif($pos[$i] eq 'V') {
	    $freq{$i}{A} += 1/3;
	    $freq{$i}{C} += 1/3;
	    $freq{$i}{G} += 1/3;
	} elsif($pos[$i] eq 'H') {
	    $freq{$i}{A} += 1/3;
	    $freq{$i}{C} += 1/3;
	    $freq{$i}{U} += 1/3;
	} elsif($pos[$i] eq 'D') {
	    $freq{$i}{A} += 1/3;
	    $freq{$i}{G} += 1/3;
	    $freq{$i}{U} += 1/3;
	} elsif($pos[$i] eq 'B') {
	    $freq{$i}{C} += 1/3;
	    $freq{$i}{G} += 1/3;
	    $freq{$i}{U} += 1/3;
	} elsif($pos[$i] eq 'S') {
	    $freq{$i}{C} += 1/2;
	    $freq{$i}{G} += 1/2;
	} elsif($pos[$i] eq 'W') {
	    $freq{$i}{A} += 1/2;
	    $freq{$i}{U} += 1/2;
	} elsif($pos[$i] eq 'K') {
	    $freq{$i}{G} += 1/2;
	    $freq{$i}{U} += 1/2;
	} elsif($pos[$i] eq 'M') {
	    $freq{$i}{A} += 1/2;
	    $freq{$i}{C} += 1/2;
	} elsif($pos[$i] eq 'Y') {
	    $freq{$i}{C} += 1/2;
	    $freq{$i}{U} += 1/2;
	} elsif($pos[$i] eq 'R') {
	    $freq{$i}{A} += 1/2;
	    $freq{$i}{G} += 1/2;
	} elsif($pos[$i] eq 'T') {
	    $freq{$i}{U}++;
	} else {
	    $freq{$i}{$pos[$i]}++;
	}
    }

}


### exclude gap - columns
if($rmGaps) {
    
    my ($max, @maxN);

    for($i=0; $i<$l; $i++) {
	$freq{$i}{A} /= $n;
	$freq{$i}{C} /= $n;
	$freq{$i}{G} /= $n;
	$freq{$i}{U} /= $n;
	$freq{$i}{X} /= $n;
	$freq{$i}{'-'} /= $n;
    }

    my ($j, %freq2);
    $j=0;
    for($i=0; $i<$l; $i++) {
	if($freq{$i}{'-'} < 1) {
	    $freq2{$j}{A}   = $freq{$i}{A};
	    $freq2{$j}{C}   = $freq{$i}{C};
	    if($freq{$i}{A} > $freq{$i}{C}) { $max = $freq{$i}{A}; $maxN[$j] = 'A'; }
	    else { $max = $freq{$i}{C}; $maxN[$j] = 'C'; } 
	    $freq2{$j}{G}   = $freq{$i}{G};
	    if($freq{$i}{G} > $max) { $max = $freq{$i}{G}; $maxN[$j] = 'G'; }
	    $freq2{$j}{U}   = $freq{$i}{U};
	    if($freq{$i}{U} > $max) { $max = $freq{$i}{U}; $maxN[$j] = 'U'; }
	    $freq2{$j}{X}   = $freq{$i}{X};
	    $freq2{$j}{'-'} = $freq{$i}{'-'};
	    if($freq{$i}{'-'} > $max) { $max = $freq{$i}{'-'}; $maxN[$j] = '-'; }
	    $j++;
	} 
    }
    
    $l = $j; # update seq length
 
    printf "#   %s    ",$maxN[0];;
    for($i=1; $i<$l; $i++) { printf "     %s    ",$maxN[$i]; } printf "\n";
    for($i=0; $i<$l; $i++) { printf "%f  ",$freq2{$i}{A}; } printf "\n";
    for($i=0; $i<$l; $i++) { printf "%f  ",$freq2{$i}{C}; } printf "\n";
    for($i=0; $i<$l; $i++) { printf "%f  ",$freq2{$i}{G}; } printf "\n";
    for($i=0; $i<$l; $i++) { printf "%f  ",$freq2{$i}{U}; } printf "\n";
    for($i=0; $i<$l; $i++) { printf "%f  ",$freq2{$i}{'-'}; } printf "\n";


} else {

    my ($max, @maxN);

    for($i=0; $i<$l; $i++) {
	$freq{$i}{A} /= $n;
	$freq{$i}{C} /= $n;
	if($freq{$i}{A} > $freq{$i}{C}) { $max = $freq{$i}{A}; $maxN[$i] = 'A'; }
	else { $max = $freq{$i}{C}; $maxN[$i] = 'C'; } 
	$freq{$i}{G} /= $n;
	if($freq{$i}{G} > $max) { $max = $freq{$i}{G}; $maxN[$i] = 'G'; }
	$freq{$i}{U} /= $n;
	if($freq{$i}{U} > $max) { $max = $freq{$i}{U}; $maxN[$i] = 'U'; }
	$freq{$i}{X} /= $n;
	$freq{$i}{'-'} /= $n;
	if($freq{$i}{'-'} > $max) { $max = $freq{$i}{'-'}; $maxN[$i] = '-'; }
    }

    printf "#   %s    ",$maxN[0];;
    for($i=1; $i<$l; $i++) { printf "     %s    ",$maxN[$i]; } printf "\n";
    for($i=0; $i<$l; $i++) { printf "%f  ",$freq{$i}{A}; } printf "\n";
    for($i=0; $i<$l; $i++) { printf "%f  ",$freq{$i}{C}; } printf "\n";
    for($i=0; $i<$l; $i++) { printf "%f  ",$freq{$i}{G}; } printf "\n";
    for($i=0; $i<$l; $i++) { printf "%f  ",$freq{$i}{U}; } printf "\n";
    for($i=0; $i<$l; $i++) { printf "%f  ",$freq{$i}{'-'}; } printf "\n";

}
