#!/usr/bin/perl -w

use strict;
use Getopt::Long;
use Pod::Usage;

my $man = 0;
my $help = 0;
my $consensus;
my ($sL, $eL, $sR, $eR) = (1, -1, 1, -1);
my ($file, $comment) = ('', '');

# ~/software/weblogo-3.0b13/weblogo -c 'classic' -A 'rna' -U 'probability' < infiles/Kt11_web.fa > infiles/Kt11_web.eps

# ---------------------------------------------------------------------------- #
# Parse options and print usage if there is a syntax error,
# or if usage was explicitly requested.
GetOptions('help|?'  => \$help,
	   man       => \$man, 
	   consensus => \$consensus,
	   'in=s'     => \$file,
	   'note=s'  => \$comment,
	   'sL=i'    => \$sL,
	   'eL=i'    => \$eL,
	   'sR=i'    => \$sR,
	   'eR=i'    => \$eR
    ) or pod2usage(2);

pod2usage(1) if $help;
pod2usage(-verbose => 2) if $man;

pod2usage("$0: Arguments missing.") if($file eq '');

$comment = $file if($comment eq '');

# ---------------------------------------------------------------------------- #
# Calculation starts here
open(F, $file) or die "%s",$file.$!; my @fasta = <F>; close(F);

# usr starts with 1, here we start with 0 !
$sL -= 1;
if($eL >= 1) { -= 1; }
$sR -= 1;
if($eR >= 1) { -= 1; }

my ($l, $n, $id, %freq, $i);
my (@freqAl, @freqCl, @freqGl, @freqUl, @gapsl);
my (@freqAr, @freqCr, @freqGr, @freqUr, @gapsr);

# length of sequences
$l = `head -2 $file | grep -v '>' | wc  |awk '{printf "%d", \$NF;}'`;
# number of entries
$n = `grep '>' $file | wc  |awk '{printf "%d", \$1;}'`;

pod2usage("$0: Infile contains 0 sequences!. Stopped.") if($n == 0);

## initialize arrays
my ($j, $nL, $nR);
$nL = $eL - $sL + 1;
$nR = $eR - $sR + 1;

for($i=0; $i<$nL; $i++) {
    $freqAl[$i] = 0; $freqCl[$i] = 0; $freqGl[$i] = 0; $freqUl[$i] = 0; $gapsl[$i] = 0;
}
for($i=0; $i<$nR; $i++) {
    $freqAr[$i] = 0; $freqCr[$i] = 0; $freqGr[$i] = 0; $freqUr[$i] = 0; $gapsr[$i] = 0;
}

## fill freq arrays
my $seq;
my $c = 0;
for($j=0; $j<scalar(@fasta); $j++) {
    if($j%2 == 1) {
	$seq = uc $fasta[$j];
	$seq =~ s/T/U/g;
	$seq =~ s/\./\-/g;
	chomp($seq);

	my @pos = split //, $seq;
	$c = 0;
	for($i=$sL; $i<=$eL; $i++) {
	    if($pos[$i] eq 'A') { $freqAl[$c]++; }
	    elsif($pos[$i] eq 'C') { $freqCl[$c]++; }
	    elsif($pos[$i] eq 'G') { $freqGl[$c]++; }
	    elsif($pos[$i] eq 'U') { $freqUl[$c]++; }
	    elsif($pos[$i] eq 'R') { $freqAl[$c] += 1/2; $freqGl[$c] += 1/2; }
	    elsif($pos[$i] eq 'Y') { $freqCl[$c] += 1/2; $freqUl[$c] += 1/2; }
	    elsif($pos[$i] eq 'S') { $freqGl[$c] += 1/2; $freqCl[$c] += 1/2; }
	    elsif($pos[$i] eq 'W') { $freqAl[$c] += 1/2; $freqUl[$c] += 1/2; }
	    elsif($pos[$i] eq 'K') { $freqGl[$c] += 1/2; $freqUl[$c] += 1/2; }
	    elsif($pos[$i] eq 'M') { $freqAl[$c] += 1/2; $freqCl[$c] += 1/2; }
	    elsif($pos[$i] eq 'B') { $freqCl[$c] += 1/3; $freqGl[$c] += 1/3; $freqUl[$c] += 1/3;}
	    elsif($pos[$i] eq 'D') { $freqAl[$c] += 1/3; $freqGl[$c] += 1/3; $freqUl[$c] += 1/3;  }
	    elsif($pos[$i] eq 'H') { $freqAl[$c] += 1/3; $freqCl[$c] += 1/3; $freqUl[$c] += 1/3;  }
	    elsif($pos[$i] eq 'V') { $freqAl[$c] += 1/3; $freqCl[$c] += 1/3; $freqGl[$c] += 1/3;  }
	    elsif($pos[$i] eq 'N') { $freqAl[$c] += 1/4; $freqCl[$c] += 1/4;
				     $freqGl[$c] += 1/4; $freqUl[$c] += 1/4; }
	    elsif($pos[$i] eq '-') { $gapsl[$c]++; }
	    $c++;
	}
	
	$c = 0;
	for($i=$sR; $i<=$eR; $i++) {
	    if($pos[$i] eq 'A') { $freqAr[$c]++; }
	    elsif($pos[$i] eq 'C') { $freqCr[$c]++; }
	    elsif($pos[$i] eq 'G') { $freqGr[$c]++; }
	    elsif($pos[$i] eq 'U') { $freqUr[$c]++; }
	    elsif($pos[$i] eq 'R') { $freqAr[$c] += 1/2; $freqGr[$c] += 1/2; }
	    elsif($pos[$i] eq 'Y') { $freqCr[$c] += 1/2; $freqUr[$c] += 1/2; }
	    elsif($pos[$i] eq 'S') { $freqGr[$c] += 1/2; $freqCr[$c] += 1/2; }
	    elsif($pos[$i] eq 'W') { $freqAr[$c] += 1/2; $freqUr[$c] += 1/2; }
	    elsif($pos[$i] eq 'K') { $freqGr[$c] += 1/2; $freqUr[$c] += 1/2; }
	    elsif($pos[$i] eq 'M') { $freqAr[$c] += 1/2; $freqCr[$c] += 1/2; }
	    elsif($pos[$i] eq 'B') { $freqCr[$c] += 1/3; $freqGr[$c] += 1/3; $freqUr[$c] += 1/3;}
	    elsif($pos[$i] eq 'D') { $freqAr[$c] += 1/3; $freqGr[$c] += 1/3; $freqUr[$c] += 1/3;  }
	    elsif($pos[$i] eq 'H') { $freqAr[$c] += 1/3; $freqCr[$c] += 1/3; $freqUr[$c] += 1/3;  }
	    elsif($pos[$i] eq 'V') { $freqAr[$c] += 1/3; $freqCr[$c] += 1/3; $freqGr[$c] += 1/3;  }
	    elsif($pos[$i] eq 'N') { $freqAr[$c] += 1/4; $freqCr[$c] += 1/4;
				     $freqGr[$c] += 1/4; $freqUr[$c] += 1/4; }
	    elsif($pos[$i] eq '-') { $gapsr[$c]++; }
	    $c++;
	}
	
    }
}

# splice gap columns
my (@splicel, @splicer);

for($i=0; $i<$nL; $i++) {
    $freqAl[$i] /= $n;
    $freqCl[$i] /= $n;
    $freqGl[$i] /= $n;
    $freqUl[$i] /= $n;
    $gapsl[$i]  /= $n;

    if($gapsl[$i] > 0.99) {
	push @splicel, $i;
    }
}

for($i=0; $i<scalar(@splicel); $i++) {
    splice @freqAl, $splicel[$i]-$i, 1;
    splice @freqCl, $splicel[$i]-$i, 1;
    splice @freqGl, $splicel[$i]-$i, 1;
    splice @freqUl, $splicel[$i]-$i, 1;
    splice @gapsl,  $splicel[$i]-$i, 1;
}

$nL = scalar @freqAl;

for($i=0; $i<$nR; $i++) {
    $freqAr[$i] /= $n;
    $freqCr[$i] /= $n;
    $freqGr[$i] /= $n;
    $freqUr[$i] /= $n;
    $gapsr[$i]  /= $n;

    if($gapsr[$i] > 0.99) {
	push @splicer, $i;
    }
}

for($i=0; $i<scalar(@splicer); $i++) {
    splice @freqAr, $splicer[$i]-$i, 1;
    splice @freqCr, $splicer[$i]-$i, 1;
    splice @freqGr, $splicer[$i]-$i, 1;
    splice @freqUr, $splicer[$i]-$i, 1;
    splice @gapsr,  $splicer[$i]-$i, 1;
}

$nR = scalar @freqAr;

# get consensus sequence
my ($max, @maxL, @maxR);

for($i=0; $i<$nL; $i++) {
    $max = 0;
    if($freqAl[$i] > $max) { $max = $freqAl[$i]; $maxL[$i] = 'A'; }
    if($freqCl[$i] > $max) { $max = $freqCl[$i]; $maxL[$i] = 'C'; }
    if($freqGl[$i] > $max) { $max = $freqGl[$i]; $maxL[$i] = 'G'; }
    if($freqUl[$i] > $max) { $max = $freqUl[$i]; $maxL[$i] = 'U'; }
}
for($i=0; $i<$nR; $i++) {
    $max = 0;
    if($freqAr[$i] > $max) { $max = $freqAr[$i]; $maxR[$i] = 'A'; }
    if($freqCr[$i] > $max) { $max = $freqCr[$i]; $maxR[$i] = 'C'; }
    if($freqGr[$i] > $max) { $max = $freqGr[$i]; $maxR[$i] = 'G'; }
    if($freqUr[$i] > $max) { $max = $freqUr[$i]; $maxR[$i] = 'U'; }
}

# print results
printf "# %s\n",$comment;
printf "%d # left stem size\n", $nL;
for($i=0; $i<$nL; $i++) { printf "%f   ", $freqAl[$i]; } printf "\n";
for($i=0; $i<$nL; $i++) { printf "%f   ", $freqCl[$i]; } printf "\n";
for($i=0; $i<$nL; $i++) { printf "%f   ", $freqGl[$i]; } printf "\n";
for($i=0; $i<$nL; $i++) { printf "%f   ", $freqUl[$i]; } printf "\n";
printf "#   %s    ",$maxL[0];
for($i=1; $i<$nL; $i++) { printf "      %s    ", $maxL[$i]; } printf "\n";

printf "%d # right stem size\n", $nR;
for($i=0; $i<$nR; $i++) { printf "%f   ", $freqAr[$i]; } printf "\n";
for($i=0; $i<$nR; $i++) { printf "%f   ", $freqCr[$i]; } printf "\n";
for($i=0; $i<$nR; $i++) { printf "%f   ", $freqGr[$i]; } printf "\n";
for($i=0; $i<$nR; $i++) { printf "%f   ", $freqUr[$i]; } printf "\n";
printf "#   %s    ",$maxR[0];
for($i=1; $i<$nR; $i++) { printf "      %s    ", $maxR[$i]; } printf "\n";


## length of left and right stem
__END__



## print results
printf "# %s\n",$comment;
printf "%d # left stem size\n", $nL;
for($i=0; $i<$nL; $i++) { printf "%f  ",$left{$i}{A}; } printf "\n";
for($i=0; $i<$nL; $i++) { printf "%f  ",$left{$i}{C}; } printf "\n";
for($i=0; $i<$nL; $i++) { printf "%f  ",$left{$i}{G}; } printf "\n";
for($i=0; $i<$nL; $i++) { printf "%f  ",$left{$i}{U}; } printf "\n";
if($consensus) {
    printf "#   %s    ",$maxN[0];;
    for($i=0; $i<$nL; $i++) { printf "     %s    ",$maxNl[$i]; } printf "\n";
}
printf "%d # right stem size\n", $nR;
for($i=0; $i<$nR; $i++) { printf "%f  ",$right{$i}{A}; } printf "\n";
for($i=0; $i<$nR; $i++) { printf "%f  ",$right{$i}{C}; } printf "\n";
for($i=0; $i<$nR; $i++) { printf "%f  ",$right{$i}{G}; } printf "\n";
for($i=0; $i<$nR; $i++) { printf "%f  ",$right{$i}{U}; } printf "\n";

if($consensus) {
    printf "#   %s    ",$maxN[0];;
    for($i=0; $i<$nR; $i++) { printf "     %s    ",$maxNr[$i]; } printf "\n";

}

printf STDERR "NOTE: Please extend the motif file by a table of basepairs and corresponding bp types\n";


# ---------------------------------------------------------------------------- #
sub max {

    my $a = shift;
    my $b = shift;

    if($a > $b) return $a;
    else return $b;
    
}

__END__

=head1 NAME

reformatMotif.pl - reformatting a sequence motif provided in FASTA format

=head1 SYNOPSIS

reformatMotif.pl -sL N -eL N -sR N -sL N -i FILE  

Options:
    -help brief help message
    -man full documentation

    [REQUIRED:]
    -sL N               Start position of left stem in motif file starting with 1
    -eL N               End position of left stem in motif file
    -sR N               Start position of right stem in motif file
    -eR N               End position of right stem in motif file
    -i FILE             FASTA formatted file of the motif

    [OPTIONAL:]
    - note "STRING"     comment your motif, if ommitted, file name is regarded as comment
    - consensus         print consensus sequence (iupac coded)

=head1 OPTIONS

=over 8

=item B<-help>

Print a brief help message and exits.

=item B<-man>

Prints the manual page and exits.

=back

=head1 DESCRIPTION

B<reformatMotif.pl> reads a 2-sided sequence motif as FASTA alignment,
calculates base frequency for each column and removes columns where
the gap frequency exceeds 99%. The motif is broken apart in the two
respective stems and their lengths as well as the frequency and
consensus sequence are returned to STDOUT.

=cut
