#include "calc.h"

extern char bases[]; /* ACGUTRYMKWSBDHVNacgutrymkwsbdhvnt */
/*
  A  C  G  U  T  R  Y  M  K  W  S  B  D  H  V  N  
 65 67 71 85 84 82 89 77 75 87 83 66 68 72 86 78
  a  c  g   u   t   r   y   m   k   w   s   b   d   h   v   n   t
 97 99 103 117 116 114 121 109 107 119 115 98  100 104 118 110 116
*/

/* conservative statistical estimation of score distribution */

/** Base pair occurrence frequencies within each base pair family. Extracted from:
    Leontis et al : NAR, 2009, Vol. 37, No. 7 - Figure 5
    compared with:
    Westhof et al : NAR, 2002, Vol. 30, No. 16 - Figure 2 - 13 
*/

/* -------------------------------------------------------------------------- */

char * gotoh_revCompl(char *sequence, int v, int len) {

  int i,j;
  char nt;
  char *result;

  result = (char*) calloc(len+1, sizeof(char));
  
  for(i=len-1,j=0;i>=0;i--,j++) {
    switch(sequence[i]) {
    case 'A' : nt = 'U' ; break;
    case 'U' : nt = 'A' ; break;
    case 'C' : nt = 'G' ; break;
    case 'G' : nt = 'C' ; break;
    case '-' : nt = '-' ; break;
    case 'R' : nt = 'Y' ; break;
    case 'Y' : nt = 'R' ; break;
    case 'S' : nt = 'S' ; break;
    case 'W' : nt = 'W' ; break;
    case 'K' : nt = 'M' ; break;
    case 'M' : nt = 'K' ; break;
    case 'B' : nt = 'V' ; break;
    case 'D' : nt = 'H' ; break;
    case 'H' : nt = 'D' ; break;
    case 'V' : nt = 'B' ; break;
    case 'N' : nt = 'N' ; break;
    default :
      gotoh_Warning("There are strange letters in your sequence. Parsed to 'N'.", v);
      nt='N' ;
    }
    result[j]=nt;
  }

  return result;
  
}

/* -------------------------------------------------------------------------- */

int * gotoh_getMaxS(int scores[], int s, int e) {

  int i, *ret=NULL;

  ret = (int *) calloc(2,sizeof(int));
  ret[0] = 0;  /* position in database */
  ret[1] = -1; /* score at position ret[0] */
  
  for(i=s; i<=e; i++)
    if(scores[i] >= ret[1]) { /* get last max (if equal) */
      ret[0] = i; ret[1] = scores[i];
    }

  return ret;  
  
}

/* -------------------------------------------------------------------------- */

float gotoh_scoreMotif(char s, float **freq, float *minf, float *maxf, float *infof, int j) {

  char bases2[] = "ACGU";
  int p=-1;
  float min, max, info, curr, score, tmp;

  if(strchr(bases2, s) != NULL) {
    p = (int) (strchr(bases2, s) - bases2);

    info = infof[j];
    min = info * minf[j];
    max = info * maxf[j]; 

    curr = info * freq[p][j];
    
    if(max == min && max != 0) score = (curr-min)/max;
    else if(max == 0) score = 0;
    else score = (curr - min) / (max - min);

  } else {
    switch(s) {
    case 'N':
      tmp = minf[j];
      break;
    case 'V': /* A, C, G */
      tmp = maxF(freq[0][j], maxF(freq[1][j], freq[2][j]));
      break;
    case 'H': /* A, C, U */
      tmp = maxF(freq[0][j], maxF(freq[1][j], freq[3][j]));
      break;
    case 'D': /* A, G, U */
      tmp = maxF(freq[0][j], maxF(freq[2][j], freq[3][j]));
      break;
    case 'B': /* A, G, U */
      tmp = maxF(freq[1][j], maxF(freq[2][j], freq[3][j]));
      break;
    case 'M': /* A, C */
      tmp = maxF(freq[0][j], freq[1][j]);
      break;
    case 'K': /* G, U */
      tmp = maxF(freq[2][j], freq[3][j]);
      break;
    case 'W': /* A, U */
      tmp = maxF(freq[0][j], freq[3][j]);
      break;
    case 'S': /* G, C */
      tmp = maxF(freq[1][j], freq[2][j]);
      break;
    case 'Y': /* C, U */
      tmp = maxF(freq[1][j], freq[3][j]);
      break;
    case 'R': /* A, G */
      tmp = maxF(freq[0][j], freq[2][j]);
      break;
    case '-': /* will probably never occur */
      tmp = freq[4][j];
    default:
      tmp = minf[j];
    }

    info = infof[j];
    min = info * minf[j];
    max = info * maxf[j]; 
    /* printf("info: %f  max: %f  min: %f\n", info, max, min); */
    
    curr = info * tmp;
    
    if(max == min) score = (curr-min)/max;
    else score = (curr - min) / (max - min);
    
  }
  
  /* printf("mot: %f  curr: %f  score: %f pos: %d\n", freq[p][j], curr, score,j); */
  
  return score;
  
}
/* -------------------------------------------------------------------------- */

/* score_distr => data, distribution of all possible scores */
/* double * gotoh_getParams(int * score_distr, int n_s, int n_q, */
/* 			 gotoh_settings gs, char * qname, int *ntf) { */
  
/*   if(gs.stat_file != NULL) { */
/*     maxN = 0; */
/*     for(i=fv.startscore; i<=fv.endscore; i++) { if(maxN < score_distr[i]) maxN = score_distr[i]; } */
/*     maxY = floor(log(maxN)) + 2; */
/*     /\* minX = floor((float)(fv.startscore - fv.startscore*0.5 - minAS) / 10) * 10; *\/ */
/*     /\*FF maxX = n_s + 20 - minAS; *\/ */
/*     maxX = n_s - minAS + 20; */
/*     ticks = floor((maxX)/10/10) * 10; */

/*     datafile = (char *) calloc(strlen(qname)+16, sizeof(char)); */
/*     strcpy(datafile,qname); */
/*     strcat(datafile, ".GotohScan.agr"); */
/*     fp_data = fopen(datafile, "w"); */
/*     fprintf(fp_data, "# Grace project file\n"); */
/*     fprintf(fp_data, "#\n"); */
/*     fprintf(fp_data, "@version 50120\n"); */
/*     fprintf(fp_data, "@page size 792, 612\n"); */
/*     fprintf(fp_data, "@map font 4 to \"Helvetica\", \"Helvetica\"\n"); */
/*     fprintf(fp_data, "@map font 5 to \"Helvetica-Oblique\", \"Helvetica-Oblique\"\n"); */
/*     fprintf(fp_data, "@with g0\n"); */
/*     /\*FF fprintf(fp_data, "@ world 0, -2, %d, %d\n",maxX,maxY); *\/ */
/*     fprintf(fp_data, "@ world 0, -2, %d, %d\n",maxX,maxY); */
/*     fprintf(fp_data, "@ stack world 0, 0, 0, 0\n"); */
/*     fprintf(fp_data, "@ znorm 1\n"); */
/*     fprintf(fp_data, "@ view 0.150000, 0.150000, 1.150000, 0.850000\n"); */
/*     fprintf(fp_data, "@ title \"GotohScan - Evalue calculation\"\n"); */
/*     fprintf(fp_data, "@ title font 4\n"); */
/*     fprintf(fp_data, "@ subtitle \"%s, theta=%f, k=%f\"\n",qname,fv.theta,fv.k); */
/*     fprintf(fp_data, "@ subtitle font 5\n"); */
/*     fprintf(fp_data, "@ yaxes scale Normal\n"); */
/*     fprintf(fp_data, "@ xaxes scale Normal\n"); */
/*     fprintf(fp_data, "@ xaxis label \"alignment score\"\n"); */
/*     fprintf(fp_data, "@ xaxis label font 4\n"); */
/*     fprintf(fp_data, "@ xaxis ticklabel font 4\n"); */
/*     fprintf(fp_data, "@ xaxis tick place both\n"); */
/*     fprintf(fp_data, "@ xaxis tick major %d\n",ticks); */
/*     fprintf(fp_data, "@target G0.S1\n"); */
/*     fprintf(fp_data, "@type xy\n"); */
/* /\*     fprintf(fp_data, "@ xaxis tick minor ticks 1\n"); *\/ */
/*     fprintf(fp_data, "@ yaxis label \"log( # alns )\"\n"); */
/*     fprintf(fp_data, "@ yaxis label font 4\n"); */
/*     fprintf(fp_data, "@ yaxis ticklabel font 4\n"); */
/*     fprintf(fp_data, "@ yaxis tick place both\n"); */
/*     fprintf(fp_data, "@ yaxis tick major 5\n"); */
/* /\*     fprintf(fp_data, "@ yaxis tick minor ticks 1\n"); *\/ */
/*     fprintf(fp_data, "@ legend on\n"); */
/*     fprintf(fp_data, "@ legend font 4\n"); */
/*     fprintf(fp_data, "@ legend char size 0.9\n"); */
/*     fprintf(fp_data, "@ s0 symbol 10\n"); */
/*     fprintf(fp_data, "@ s0 symbol size 0.160000\n"); */
/*     fprintf(fp_data, "@ s0 line type 0\n"); */
/*     fprintf(fp_data, "@ s0 legend \"Score distribution [%d,%d]\"\n",-minAS,n_s-minAS-1); */
/*     fprintf(fp_data, "@ s1 line linewidth 1.5\n"); */
/*     fprintf(fp_data, "@ s1 line color 2\n"); */
/*     fprintf(fp_data, "@ s1 line style 2\n"); */
/*     fprintf(fp_data, "@ s1 legend \"Fitted gamma distribution\"\n"); */

/*     fprintf(fp_data, "@ s2 line type 1\n"); */
/*     fprintf(fp_data, "@ s2 line style 3\n"); */
/*     fprintf(fp_data, "@ s2 line linewidth 1.0\n"); */
/*     fprintf(fp_data, "@ s2 line color 7\n"); */
/*     fprintf(fp_data, "@ s2 line pattern 1\n"); */
/*     fprintf(fp_data, "@ s2 legend \"Portion used for fitting\"\n"); */

/*     fprintf(fp_data, "@ s3 line type 1\n"); */
/*     fprintf(fp_data, "@ s3 line style 3\n"); */
/*     fprintf(fp_data, "@ s3 line linewidth 1.0\n"); */
/*     fprintf(fp_data, "@ s3 line color 7\n"); */
/*     fprintf(fp_data, "@ s3 line pattern 1\n"); */

/*     fprintf(fp_data, "@target G0.S0\n"); */
/*     fprintf(fp_data, "@type xy\n"); */

/*     /\* print data points *\/ */
/*     for(i=0;i<=n_s;i++) { */
/*       if(score_distr[i] <= 0) y = -1; */
/*       else y = log(score_distr[i]); */
/*       x = i - minAS; */
/*       logx=log(i)/log(minAS); */
/*       fprintf(fp_data, "%d %lf\n", x, y); */
/*       fprintf(stdout, "S %10d %10d %lf   ",x,score_distr[i],y); */
/*       y = log(fv.N) + fv.k*logx - logx - x/fv.theta - fv.k*log(fv.theta) - gamma; */
/*       fprintf(stdout, "%lf\n",y); */
/*     } */

/*     fprintf(fp_data, "&\n"); /\* seperate data from fitted curve *\/ */

/*     /\* print fitted gamma curve *\/ */
/*     for(i=fv.startscore;i<=n_s;i++) { */
/*       x=i-minAS; */
/*       logx=log(i)/log(minAS); */
/*       y = log(fv.N) + fv.k*logx - logx - x/fv.theta - fv.k*log(fv.theta) - gamma; */
/*       /\*FF fprintf(fp_data, "%d %lf\n", x-minAS, y);     *\/ */
/*       fprintf(fp_data, "%d %lf\n", x, y);     */
/*     } */

/*     fprintf(fp_data, "&\n"); */

/*     fprintf(fp_data, "%d\t%d\n",fv.startscore-minAS,maxX); */
/*     fprintf(fp_data, "%d\t-3\n",fv.startscore-minAS); */
/*     fprintf(fp_data, "&\n"); */

/*     fprintf(fp_data, "%d\t%d\n",fv.endscore-minAS,maxX); */
/*     fprintf(fp_data, "%d\t-3\n",fv.endscore-minAS); */
/*     fprintf(fp_data, "&\n"); */


/*     fflush(fp_data); */
/*     fclose(fp_data); */
/*     free(datafile); */
/*   } */
  
 /*  exit(0); */
/*   params = (double *) calloc(4, sizeof(double)); */
/*   params[0] = (double)fv.k; */
/*   params[1] = (double)fv.theta; */
/*   params[2] = (double)gamma; */
/*   params[3] = (double)fv.N; */
  
/*   return params; */
  
/* } */

/* -------------------------------------------------------------------------- */

struct gotoh_aln gotoh_initL(void) {

  gotoh_aln ga;

  ga.db_len = 0;
  ga.hitseq=NULL; ga.queryseq=NULL; ga.hitname=NULL; ga.queryname=NULL;
  ga.sStart = -1; ga.sEnd = -1;
  ga.qStart = -1; ga.qEnd = -1;
  ga.length = 0;
  ga.alnscore = -1;
  ga.ntfalnscore = -1;
  ga.evalue = 1e-3;
  ga.strand = '+';
  ga.mismatches = 0; ga.gapopening = 0;
  ga.percID = 0;
  
  return ga;
  
}

/* -------------------------------------------------------------------------- */

void gotoh_freeL(struct gotoh_aln a) {

  if(a.hitseq)    free(a.hitseq);
  if(a.queryseq)  free(a.queryseq);
  if(a.hitname)   free(a.hitname);
  if(a.queryname) free(a.queryname);
  
}
/* -------------------------------------------------------------------------- */

void gotoh_printGotohAln(gotoh_aln aln, int f, FILE *outfile) {

  int i;

  switch(f) {
  case 0: /* blast */
    fprintf(outfile, "%s\t%s\t%5.2f\t",aln.queryname,aln.hitname,aln.percID);
    fprintf(outfile, "%d\t%d\t%d\t",aln.length,aln.mismatches,aln.gapopening);
    if(aln.strand == '+') {
      fprintf(outfile, "%d\t%d\t%d\t%d\t",aln.qStart,aln.qEnd,aln.sStart,aln.sEnd);
    } else {
      fprintf(outfile, "%d\t%d\t%d\t%d\t",aln.qStart,aln.qEnd,aln.sEnd,aln.sStart);
    }
    fprintf(outfile, "%e\t",aln.evalue);
    fprintf(outfile, "%.2f\t", aln.ntfalnscore);
    fprintf(outfile, "%c\n", aln.strand);
    break;
  case 1: /* blast + aln sequences */
    fprintf(outfile, "%s\t%s\t%5.2f\t",aln.queryname,aln.hitname,aln.percID);
    fprintf(outfile, "%d\t%d\t%d\t",aln.length,aln.mismatches,aln.gapopening);
    if(aln.strand == '+') {
      fprintf(outfile, "%d\t%d\t%d\t%d\t",aln.qStart,aln.qEnd,aln.sStart,aln.sEnd);
    } else {
      fprintf(outfile, "%d\t%d\t%d\t%d\t",aln.qStart,aln.qEnd,aln.sEnd,aln.sStart);
    }
    fprintf(outfile, "%e\t",aln.evalue);
    fprintf(outfile, "%.2f\t", aln.ntfalnscore);
    fprintf(outfile, "strand:%c\t", aln.strand);
    fprintf(outfile, "%s&%s\n",aln.hitseq,aln.queryseq); 
    break;
  case 2: /* fasta */
    fprintf(outfile, ">%s:%s:%d-%d:%c:%.2f:%e\n",aln.queryname,aln.hitname,aln.sStart,aln.sEnd,aln.strand,aln.percID,aln.evalue);
    for(i=0; i<aln.length; i++)
      if(aln.hitseq[i] != '-') fprintf(outfile, "%c",aln.hitseq[i]);
    fprintf(outfile, "\n");
    break;
  case 3: /* maf */
    /* give real sequence lengths and NOT alignment lengths! */
    { int i,ll;
      char *t=NULL;
      t = (char *) calloc(32, sizeof(char));
      strncpy(t, aln.hitname, 30);
      
      fprintf(outfile, "a score=%e\n",aln.evalue);
      fprintf(outfile, "%-30s %13d %5d %c %13d %s\n",t,aln.sStart,aln.length,aln.strand,aln.db_len,aln.hitseq);
      strncpy(t, aln.queryname, 30);
      ll = 0;
      for(i=0; i<aln.length; i++)
	if(aln.queryseq[i] != '-')
	  ll++;
      fprintf(outfile, "%-30s %13d %5d %c %13d %s\n",t,aln.qStart,ll,aln.strand,ll,aln.queryseq);
      free(t);
    }
    fprintf(outfile, "\n");
    break;
  case 4: /* bed */
    fprintf(outfile, "%s\t%d\t%d\t%s\t%e\t%c\t",aln.hitname,aln.sStart,aln.sEnd,aln.queryname,aln.evalue,aln.strand);
    fprintf(outfile, "note=\"%s&%s\"\n",aln.hitseq,aln.queryseq);
    break;
  case 5: /* gff */
    fprintf(outfile, "%s\tGotohScan\t%s\t%d\t%d\t",aln.hitname,aln.queryname,aln.sStart,aln.sEnd);
    fprintf(outfile, "%e\t%c\t",aln.evalue,aln.strand);
    fprintf(outfile, "note=\"%s&%s\"\n",aln.hitseq,aln.queryseq);
    break;
  case 6: /* my own test output */
    fprintf(outfile, "%s\n%s\n",aln.hitname,aln.hitseq);
    fprintf(outfile, "%s\n%s\n",aln.queryname,aln.queryseq);
    fprintf(outfile, "start: %d length: %d strand: %c ",aln.sStart,aln.length,aln.strand);
    fprintf(outfile, "alnscore: %d Evalue: %.2e\n\n",(int)(aln.alnscore+0.5),aln.evalue);
    break;
  default: /* blast -> 0 */
    fprintf(outfile, "%s\t%s\t%5.2f\t",aln.queryname,aln.hitname,aln.percID);
    fprintf(outfile, "%d\t%d\t%d\t",aln.length,aln.mismatches,aln.gapopening);
    if(aln.strand == '+') {
      fprintf(outfile, "%d\t%d\t%d\t%d\t",aln.qStart,aln.qEnd,aln.sStart,aln.sEnd);
    } else {
      fprintf(outfile, "%d\t%d\t%d\t%d\t",aln.qStart,aln.qEnd,aln.sEnd,aln.sStart);
    }
    fprintf(outfile, "%e\t",aln.evalue);
    fprintf(outfile, "%.2f\t", aln.ntfalnscore);
    fprintf(outfile, "%c\n", aln.strand);
  }

  fflush(NULL);
  return;
  
}


/* -------------------------------------------------------------------------- */

void gotoh_printMotifCandidate(gotoh_motifCandidate c, char *db_name, gotoh_motif m, int f) {

  switch(f) {
  case 0:
    fprintf(stdout, "%-16s %s\n","DB-name:", db_name);
    fprintf(stdout, "%-16s %s\n","Motif:",m.name);
    fprintf(stdout, "%-16s left: %d %d right: %d %d\n","Position:",c.sL,c.eL,c.sR,c.eR);
    fprintf(stdout, "%-16s %d\n", "Distance:", c.distance);
    fprintf(stdout, "%-16s ", "Single scores:");
    fprintf(stdout, "left: %f (%5.2f%%) ", c.scoreL,c.scoreL_perc);
    fprintf(stdout, "right: %f (%5.2f%%) ",c.scoreR,c.scoreR_perc);
    fprintf(stdout, "basepairs: %f (%5.2f%%)\n", c.scoreBP, c.scoreBP_perc);
    fprintf(stdout, "%-16s complete: %f (%5.2f%%)\n", "", c.score, c.score_perc);
    fprintf(stdout, "DB %s [..] %s\n",c.left_seq, c.right_seq);
    fprintf(stdout, "M  %s [..] %s\n\n",c.left_mseq, c.right_mseq);
    break;
  case 4: /* print motif as vector */
    fprintf(stdout, "%f\t %f\t ",c.score, c.score_perc);
    fprintf(stdout, "%d...%d\t %d...%d\t ",c.sL,c.eL,c.sR,c.eR);
    fprintf(stdout, "%f\t %f\t %f\t %f\t ",c.scoreL,c.scoreL_perc,c.scoreR,c.scoreR_perc);
    fprintf(stdout, "%f\t %f\t ",c.scoreBP, c.scoreBP_perc);
    fprintf(stdout, "%s...%s\t %s...%s\t ", c.left_seq, c.right_seq, c.left_mseq, c.right_mseq);
    fprintf(stdout, "%s %s %d ",db_name,m.name,c.distance);
    fprintf(stdout, "\n");
    break;
  default:
    fprintf(stdout, "%-16s %s\n","DB-name:", db_name);
    fprintf(stdout, "%-16s %s\n","Motif:",m.name);
    fprintf(stdout, "%-16s left: %d %d right: %d %d\n","Position:",c.sL,c.eL,c.sR,c.eR);
    fprintf(stdout, "%-16s %d\n", "Distance:", c.distance);
    fprintf(stdout, "%-16s ", "Single scores:");
    fprintf(stdout, "left: %f (%5.2f%%) ", c.scoreL,c.scoreL_perc);
    fprintf(stdout, "right: %f (%5.2f%%) ",c.scoreR,c.scoreR_perc);
    fprintf(stdout, "basepairs: %f (%5.2f%%)\n", c.scoreBP, c.scoreBP_perc);
    fprintf(stdout, "%-16s complete: %f (%5.2f%%)\n", "", c.score, c.score_perc);
    fprintf(stdout, "DB %s [..] %s\n",c.left_seq, c.right_seq);
    fprintf(stdout, "M  %s [..] %s\n\n",c.left_mseq, c.right_mseq);
  }
 
}

/* -------------------------------------------------------------------------- */

float maxF(float a, float b) {

  if(a>b) return a;
  else return b;
  
}

/* -------------------------------------------------------------------------- */

int maxI(int a, int b) {

  if(a>b) return a;
  else return b;
  
}
/* -------------------------------------------------------------------------- */

struct iarray gotoh_initArrayI(int size) {

  struct iarray a;
  int i;

  a.size = size;

  a.vals  = (int *) calloc(size, sizeof(int));
  a.start = (int *) calloc(size, sizeof(int));
  
  for(i=0; i<size; i++) { a.vals[i] = 0; a.start[i] = 0; }
  
  return a;
  
}

/* -------------------------------------------------------------------------- */

void gotoh_freeArrayI(struct iarray a) {

  free(a.vals);
  free(a.start);
  
  return;

}

/* -------------------------------------------------------------------------- */

struct farray gotoh_initArrayF(int size) {

  struct farray a;
  int i;

  a.size = size;

  a.vals = (float *) calloc(size, sizeof(float));
  a.start = (int *) calloc(size, sizeof(int));
  
  for(i=0; i<size; i++) { a.vals[i] = 0.; a.start[i] = 0; }
  
  return a;
  
}

/* -------------------------------------------------------------------------- */

void gotoh_freeArrayF(struct farray a) {

  free(a.vals);
  free(a.start);

  return;

}

/* -------------------------------------------------------------------------- */

struct darray gotoh_initArrayD(int size) {

  struct darray a;
  int i;

  a.size = size;

  a.vals = (double *) calloc(size, sizeof(double));
  
  for(i=0; i<size; i++) { a.vals[i] = 0; }
  
  return a;
  
}

/* -------------------------------------------------------------------------- */

void gotoh_freeArrayD(struct darray a) {

  free(a.vals);

  return;

}

/* -------------------------------------------------------------------------- */

void gotoh_freeGotohAln(struct gotoh_aln a) {

  free(a.hitseq);
  free(a.queryseq);
  free(a.hitname);
  free(a.queryname);
  
}

/* -------------------------------------------------------------------------- */

void gotoh_copyGotohAln(struct gotoh_aln *dest, struct gotoh_aln src) {


  dest->db_len = src.db_len;
  dest->hitseq = (char *) calloc(strlen(src.hitseq)+1, sizeof(char));
  strcpy(dest->hitseq, src.hitseq);
  dest->queryseq = (char *) calloc(strlen(src.queryseq)+1, sizeof(char));
  strcpy(dest->queryseq, src.queryseq);
  dest->hitname = (char *) calloc(strlen(src.hitname)+1, sizeof(char));
  strcpy(dest->hitname, src.hitname);
  dest->queryname = (char *) calloc(strlen(src.queryname)+1, sizeof(char));
  strcpy(dest->queryname, src.queryname);
  dest->shift = src.shift;
  dest->sStart = src.sStart;
  dest->sEnd = src.sEnd;
  dest->length = src.length;
  dest->strand = src.strand;
  dest->alnscore = src.alnscore;
  dest->ntfalnscore = src.ntfalnscore;
  dest->evalue = src.evalue;
  dest->percID = src.percID;
  dest->mismatches = src.mismatches;
  dest->gapopening = src.gapopening;
  dest->qStart = src.qStart;
  dest->qEnd = src.qEnd;

}

/* -------------------------------------------------------------------------- */

float gotoh_round2(float val) {

  return floorf(val * 100);

}
