/** \file calc.h
    \brief Functions that are responsible for general
  calculation, initializations and memory allociation of certain
  structs.
    
   This file contains functions that are not specific to either
   pattern or single sequence search, rather these functions concern
   general calculations and memory handling of the GotohScan structs.

abgestuerzt bei   Helicobacter_pylori_G27/NC_011334.fna

*/

#ifndef CALC_H
#define CALC_H

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "read.h"
#include "structs.h"

#define MAX_ALNS 1000   /* max. number of alignments per query sequence */

/* alignement functions */

char * gotoh_revCompl(char *sequence, int v, int len);

int * gotoh_getMaxS(int scores[], int s, int e);

float gotoh_scoreMotif(char s, float **freq, float *min, float *max, float *info, int j);

void gotoh_backtraceAln(int ii, int imax, int smax, char strand, char *s_name, char* q_name,
			gotoh_aln List[], int m, char *db_seq, char *q_seq,
			gotoh_settings gs);

struct gotoh_fitVals gotoh_getRange (int *s, int n_s, int shift);

float gotoh_getSum(int *scores, struct gotoh_fitVals fv, float k, float theta, int shift);

void gotoh_getPercRange(int *s, int n_s, int * start, int *end, int minAS, char * qname, int *maxC, int *maxI);

struct gotoh_aln gotoh_initL(void);

void gotoh_freeL(struct gotoh_aln a);

/* utility functions */
void gotoh_printGotohAln(gotoh_aln aln, int f, FILE *outfile);

void gotoh_printMotifCandidate(gotoh_motifCandidate c, char *db_name, gotoh_motif m, int f);

float maxF(float a, float b);

int maxI(int a, int b);

/* initializes an intarray with size, returns struct containing array and size */
struct iarray gotoh_initArrayI(int size);

void gotoh_freeArrayI(struct iarray a);

struct farray gotoh_initArrayF(int size);

void gotoh_freeArrayF(struct farray a);

struct darray gotoh_initArrayD(int size);

void gotoh_freeArrayD(struct darray a);

void gotoh_freeGotohAln(struct gotoh_aln a);

void gotoh_copyGotohAln(struct gotoh_aln *dest, struct gotoh_aln src);

float gotoh_round2(float val);

#endif
