/* evalueCALC.c */
#include "evalueCALC.h"

/* -------------------------------------------------------------------------- */

int gotoh_getFitRange(struct darray x, struct darray y, int *s0, int *cut, double *slope, double *intercept) {

  double maxY=0, sumY=0;
  int i, maxI, maxX;

  /* fprintf(stderr,"logdata size: %d\n", x.size); */
  if(x.size < 1) {
    fprintf(stderr, "ERROR: array storing local maxima is empty\n");
    exit(0);
  }

  maxI = 0; maxY = 0;
  for(i=0; i<x.size; i++) {
    /* if(y.vals[i] > DBL_MAX_EXP) sumY += exp(DBL_MAX_EXP); */
    /* else  */sumY += exp(y.vals[i]);
    if(y.vals[i] > maxY) {
      maxY = y.vals[i];
      maxI = i;
    }
  }
  maxX = x.vals[maxI];

  if(sumY < 2*x.size) { /* no evalue calculation possible -> too few data points for fitting */
    return 0;
  }

  double halfY = 0.5 * maxY;
  int halfI;

  for(i=maxI; i<x.size; i++) {
    if(y.vals[i] < halfY) break;
    halfI = i;
  }
  halfY = y.vals[halfI];   /* true datapoint */
  
  float range = 0.125, lcutY, rcutY;
  int lcutI, rcutI;

  lcutY = halfY + range*maxY;
  lcutI = halfI-1;
  for(i=maxI; i<halfI; i++) {
    if(y.vals[i] < lcutY) break;
    lcutI = i;
  }
  lcutY = y.vals[lcutI];

  rcutY = halfY - range*maxY;
  rcutI = halfI+1;
  for(i=halfI+1; i<x.size; i++) {
    if(y.vals[i] < rcutY) break;
    rcutI = i;
  }
  rcutY = y.vals[rcutI];

  int lm_size = rcutI - lcutI + 1;
  double *x2=NULL, *y2=NULL;
  
  x2 = (double *) calloc(lm_size + 1, sizeof(double));
  y2 = (double *) calloc(lm_size + 1, sizeof(double));
  for(i=0; i<lm_size; i++) {
    y2[i] = y.vals[lcutI+i];
    x2[i] = x.vals[lcutI+i];
  }

  
  /* printf("## Tangente\n"); */
  /* printf("max:     %d %e\n", maxX, maxY); */
  /* printf("1/2 max: %d %e\n", (int)(x.vals[(int)(halfI+0.5)]), halfY); */
  /* printf("lcut:    %d %e\n", (int)(x.vals[(int)(lcutI+0.5)]), lcutY); */
  /* printf("rcut:    %d %e\n", (int)(x.vals[(int)(rcutI+0.5)]), rcutY); */
  /* printf("------------\n"); */
  
  double A=0, B=0, C=0, D=0;
  int S;
  
  for(i=0; i<lm_size; i++) {
    A += x2[i];
    B += y2[i];
    C += x2[i]*x2[i];
    D += x2[i]*y2[i];
  }
  S = lm_size;

  /* printf("A: %e\n", A); */
  /* printf("B: %e\n", B); */
  /* printf("C: %e\n", C); */
  /* printf("D: %e\n", D); */

  double m, N;
  m = (S*D - A*B) / (S*C - A*A);
  N = (D - m*C) / A;

  *slope = m;
  *intercept = N;
  /* printf("y=%ex + %e\n", m, N); */

  double ci;
  
  ci = (-1* N)/m;
  i=halfI;
  while(x.vals[i] <= ci) {
    if(i >= x.size) break;
    i++;
  }
  
  /* for(i=halfI; i<x.size; i++)  */
  /*   if(x.vals[i] > ci) break; */
  
  *cut = i - 1;

  int s0i;

  s0i = (int) ((x.vals[halfI] - maxX + 1.5) / 2 + maxX);
  for(i=maxI; i<halfI; i++)
    if(x.vals[i] > s0i) break;

  *s0 = i - 1;

  /* fprintf(stderr, "cut: %d s0: %d\n", cut, s0); */
  
  free(x2); free(y2);
  return 1;
  
}

/* -------------------------------------------------------------------------- */

int gotoh_calcEvalue(struct gotoh_aln L[], int sizeL, struct darray x, struct darray y,
		      int s0, int cut, int maxPossScore) {

  int i, n;
  double *x2=NULL, *y2=NULL, sumY=0;

  n = cut - s0 + 1; /* what if n == 0 ? */
  
  x2 = (double *) calloc(n, sizeof(double));
  y2 = (double *) calloc(n, sizeof(double));
  for(i=0; i<n; i++) {
    x2[i] = x.vals[s0+i];
    y2[i] = y.vals[s0+i];
    /* printf("FR %e %e\n",x2[i],y2[i]); */
    sumY += y2[i];
  }

  if(n <= 3) {

    for(i=0; i<sizeL; i++) { /* no fit possible, evalue is average distance to all other points starting at s0 */
      float x1;

      /* if alnscore < s0, return E(s0) */

      if(L[i].alnscore == -1) break;
      x1 = L[i].alnscore;

      int j;
      float sum=0, p0x, p0y;
      for(j=0; j<x.size; j++)
	if(x.vals[j] == x1)
	  break;

      p0x = x1;
      /* if(y.vals[j] > DBL_MAX_EXP) p0y = exp(DBL_MAX_EXP); */
      /* else  */p0y = exp(y.vals[j]);

      for(j=0; j<x.size; j++) {
	/* if(y.vals[j] > DBL_MAX_EXP) sum += (p0x - x.vals[j])*(p0y - exp(DBL_MAX_EXP)); */
	/* else  */sum += (p0x - x.vals[j])*(p0y - exp(y.vals[j]));
      }

      /* if(sum/x.size > DBL_MAX_EXP) L[i].evalue = exp(DBL_MAX_EXP); */
      /* else  */L[i].evalue =  exp(sum / x.size);

    }  
    fprintf(stderr, "# NOTE: Too few datapoints -> no fit possible, evalue is average distance to all other points\n");
    return 1;

  }
  
  double A=0, B=0, C=0, D=0, E=0, F=0, G=0, H=0;
  int S;

  for(i=0; i<n; i++) {
    A += x2[i];
    B += log(x2[i]);                   /* x[i] always > 0! (= scores, see above)*/
    C += x2[i]*log(x2[i]);
    D += x2[i]*x2[i];
    E += log(x2[i])*log(x2[i]);
    F += x2[i]*y2[i];
    G += y2[i]*log(x2[i]);
    H += y2[i];
  }
  S = n;

  /* printf("A= %f\nB= %f\nC= %f\nD= %f\nE= %f\nF= %f\nG= %f\nH= %f\nS= %d\n", A,B,C,D,E,F,G,H,S); */
  free(x2); free(y2);

  double k, a0, a1, a2;

  a2 = ((A*H-S*F)*(A*B-S*C) - (B*H-S*G)*(A*A-S*D)) / ((A*B-S*C)*(A*B-S*C) - (A*A-S*D)*(B*B-S*E));
  a1 = (A*H - S*F - (A*B-S*C)*a2) / (A*A-S*D);
  a0 = (H - B*a2 - A*a1) / S;

  if(a1 > 0) { /* too few datapoints to do a correct fit - fit only linear*/

    return 0;
    
  } else { /* indeed do an evalue calculation */
   
    /* fprintf(stderr, "a0: %f   a1: %f   a2: %f\n", a0, a1, a2); */
    double psi;

    psi = gotoh_psi(x.vals[cut], a0, a1, a2);
    /* if(psi > DBL_MAX_EXP) k = 1/exp(DBL_MAX_EXP); */
    /* else  */k = 1/exp(psi);

    /* fprintf(stderr, "k: %f\n",k); */
    /* fprintf(stderr, "DBL_MAX_EXP: %d\t DBL_MIN_EXP: %d\n",DBL_MAX_EXP, DBL_MIN_EXP); */
  
    for(i=0; i<sizeL; i++) {
      float x;
      double E, psi;
      if(L[i].alnscore == -1) break;
      x = L[i].alnscore;

      psi = gotoh_psi(x, a0, a1, a2);
      /* fprintf(stderr, "psi: %e\n",psi); */
      /* if(psi > DBL_MAX_EXP) E = k*exp(DBL_MAX_EXP-1); */
      /* if(psi < DBL_MIN_EXP) E =  */
      /* else  */E = k*exp(psi);

      L[i].evalue = E;
    }
    return 1;
  }
}

/* -------------------------------------------------------------------------- */

float gotoh_psi(float score, double a0, double a1, double a2) {

  float psi, i;

  i = score + 1;

  psi = a0 + a1*i + a2*log(i);
  
  return psi;

}

/* -------------------------------------------------------------------------- */


/* End of file */
