/* evalueCALC.h */

#ifndef eCALC_H
#define eCALC_H

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include "structs.h"

/**
   \brief Determine the fit range.

   \param data Distribution of local maxima. 2 Dimensions: 0->xvals, 1->yvals
   \param n Size of Distribution.
   \param s0 Left border.
   \param cut Right border.

   Determine the left and right borders of the tail of the data
   that should be used for fitting.

   \return 0 if no evalue calculation possible, 1 otherwise
 */
int gotoh_getFitRange(struct darray xvals, struct darray yvals, int *s0, int *cut, double *slope, double *intercept);

/**
   \brief Fit the psi funktion to data points.

   \param L List of local maxima. Complete backtraced alignments.
   \parma sizeL Size of List L.
   \param logdata Distribution of local maxima. 2 Dimensions: 0->xvals, 1->yvals
   \param sizeD Size of Distribution.
   \param s0 Left border.
   \param cut Right border.
   \param maxPossScore Highest possible Alignment score.

   Calculate the parameters of the Psi-funktion that fit best to the
   datapoints. Based on these parameters, determine the E-value and
   update its value in the list of alignment candidates.
   
   
 */
int gotoh_calcEvalue(struct gotoh_aln L[], int sizeL, struct darray x, struct darray y,
		      int s0, int cut, int maxPossScore);

/**
   \brief Funktion psi.

   \param score Argument of Psi-funktion for which the value should be calculated.
   \param a0 1st parameter of Psi-funktion
   \param a1 2nd parameter of Psi-funktion
   \param a2 3rd parameter of Psi-funktion

   Determine the value of the Psi-funktion given an argument and the
   estimated parameters.
   
   \return Value of Psi-funktion.
 */
float gotoh_psi(float score, double a0, double a1, double a2);


#endif

/* End of file */
