#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "read.h"
#include "singleSequenceCALC.h"
#include "motifCALC.h"
#include "evalueCALC.h"
#include "calc.h"
#include "fasta.h"


char bases[33] = "ACGUTRYMKWSBDHVNacgutrymkwsbdhvnt";
int gotoh_Table[128][128];
float gotoh_bp[13][496];

int main(int argc, char *argv[]) {

  struct gotoh_arguments arg;
  struct gotoh_settings cfg;
  struct gotoh_aln L[MAX_ALNS];   /* list of best MAX_ALNS alignments per query */
  struct gotoh_motif motif;
  struct gotoh_motifSingle mS;

  int i, k, m, N=0, lo=0, up=0, lo_transf=0, up_transf=0, size=0;
  int q_len=0, db_len=0, startx=0, cut=0;
  int *score_distr=NULL, *score_histoL=NULL, *score_histoR=NULL, sizeL=0, sizeR=0;
  char *q_seq=NULL, *q_name=NULL, *db_seq=NULL, *db_name=NULL;
  FASTAFILE *fp_db=NULL, *fp_q=NULL;
  FILE *res_fp=NULL;
  double * p=NULL;
  float *score_distrL=NULL, *score_distrR=NULL;
  
  arg = gotoh_readArguments(argc, argv);  /* gotoh_printArguments(arg);  exit(0);  */

  cfg = gotoh_readSettings(arg);  /* gotoh_printSettings(cfg); exit(0); */
  
  if(arg.m1_file) goto motif1; /* single motif search */
  
  if(arg.m2_file) goto motif2; /* only for 2-sided motifs!!! */

  fp_q = OpenFASTA(cfg.query);

  if(fp_q==NULL) {
    fprintf(stderr, "GS-ERROR: Cannot access query file: '%s'.\n",cfg.query);
    gotoh_freeArguments(arg);
    gotoh_freeSettings(cfg);    
    exit(0);
  }
      
  gotoh_initMatchTable(cfg);

  struct iarray scores_db;
  struct iarray lmScores;

  while(ReadFASTA(fp_q, &q_seq, &q_name, &q_len)) { /* for each query sequence ... */

    /* init a candidates array */
    for(i=0; i<MAX_ALNS; i++) L[i] = gotoh_initL();

    /* init a score array */
    lo = 2 * (cfg.gap_open + ((q_len - 1) * cfg.gap_extend));
    up = cfg.match * q_len;
    size = abs(lo) + abs(up) + 1;
    /* fprintf(stderr, "shift: %d qu_len: %d\n%s\n",lo,q_len,q_seq); */
    lmScores = gotoh_initArrayI(size);
      
    /* distribution of all possible scores per query */
    score_distr = (int *) calloc(size + 1, sizeof(int));      /* all scores */
    for(i=0; i<size; i++) score_distr[i] = 0;

    m = 2*q_len;

    fp_db = OpenFASTA(cfg.dbase); 
    if(fp_db == NULL) {
      fprintf(stderr, "GS-ERROR: Cannot access database file: '%s'.\n",cfg.dbase);
      exit(0);
    }

    while(ReadFASTA(fp_db, &db_seq, &db_name, &db_len)) {

      scores_db = gotoh_initArrayI(db_len+1);
      
      char *rev_db_seq=NULL;
      rev_db_seq = gotoh_revCompl(db_seq, cfg.verbose, db_len);
     
      if(db_len < q_len) {
	if(cfg.verbose == 1) {
	  fprintf(stderr, "GS-WARNING: Database sequence shorter than query.\n");
	  fprintf(stderr, "        DB: %s length: %d\n",db_name,db_len);
	  fprintf(stderr, "     QUERY: %s length: %d\n",q_name,q_len);
	  fprintf(stderr, "...skipping %s.\n",db_name);
	}
	free(db_seq);
	free(db_name);
	continue;
      } 
      if(cfg.quiet == 0) { fprintf(stderr, "# Q: %s ==> DB: %s\n",q_name,db_name); fflush(stderr); }

      gotoh_fwRecursion(db_seq, db_len, q_seq, q_len, &scores_db, score_distr, cfg, '+');

      /* insert maxima in L */
      if(cfg.quiet == 0) { fprintf(stderr, "\n#    Backtracing k best candidates "); fflush(stderr); }
      int o;
      /* for(o=0; o<scores_db.size; o++) printf("O+ %d %d\n",o, scores_db.vals[o]); */
      /* printf("dbsize: %d\n",scores_db.size); */
      /* for(o=0; o<scores_db.size; o++) */
      /* 	printf("O+ %d %d start: %d gaps: %d\n",o, scores_db.vals[o],scores_db.start[o],scores_db.gaps[o]); */
      
      gotoh_findLocalMaxima(&scores_db, L, 0, db_len-1, db_seq, db_name, q_seq, q_name, q_len, '+', cfg, &lmScores, db_len);
      fprintf(stderr, "\n");fflush(stderr);
      /* exit(0); */
      /* search revcompl */
      gotoh_fwRecursion(rev_db_seq, db_len, q_seq, q_len, &scores_db, score_distr, cfg, '-');
 
      /* insert maxima in L */
      if(cfg.quiet == 0) { fprintf(stderr, "\n#    Backtracing k best candidates "); fflush(stderr); }
      /* for(o=0; o<scores_db.size; o++) printf("O- %d %.2f\n",o, scores_db.vals[o]); */

      gotoh_findLocalMaxima(&scores_db, L, 0, db_len-1, rev_db_seq, db_name, q_seq, q_name, q_len, '-', cfg, &lmScores, db_len);
      fprintf(stderr, "\n");fflush(stderr);
      
      N++; 
  
      /* next database sequence -> */
      gotoh_freeArrayI(scores_db);
      free(db_name);
      free(db_seq);
      free(rev_db_seq);
    }
    /* fclose(fp_db); */
    /* CloseFASTA(fp_db); */

    /* rewind(fp_db); */
  
    /* for(i=0;i<lmScores.size;i++) printf("LMH %d %d   (%s)\n",i, lmScores.vals[i], q_name); */

    /* for(i=0;i<MAX_ALNS;i++) printf("MA %d %.2f %.2f %.2f\n",i,L[i].alnscore,L[i].ntfalnscore,L[i].percID); */

    if(N > 0) {

      if(cfg.res_file != NULL)
	res_fp = fopen(cfg.res_file, "w");
      else
	res_fp = stdout;
    
      struct darray x;
      struct darray y;
      
      x = gotoh_initArrayD(size);
      y = gotoh_initArrayD(size);

      k=0;
      for(i=0; i<size; i++) {
	if(lmScores.vals[i] == 0) continue;     /* zero scores are removed */
	x.vals[k] = (double)(i);
	y.vals[k] = (double)(log(lmScores.vals[i]));
	k++;
      }

      x.size = k;
      y.size = k;

      int ecalc=0;
      double slope, intercept;
      
      ecalc = gotoh_getFitRange(x, y, &startx, &cut, &slope, &intercept);

      if(ecalc == 0) {
      	if(cfg.quiet == 0) {
	  fprintf(res_fp, "# Too few alignments to estimate evalue-calculation\n");
	  fprintf(res_fp, "# Returning best 5 canidates - which might not be significant!\n");
	}
	for(i=0; i<5; i++) gotoh_printGotohAln(L[i], cfg.output_format, res_fp);
      	goto end0;
      }
      
      ecalc = gotoh_calcEvalue(L, MAX_ALNS, x, y, startx, cut, size);
      if(ecalc == 0) {
	/* too few datapoints in fit range to force the parabel falling */
	/* take tangente to get E */
	double k, fit;
	if(cfg.quiet == 0) {
	  fprintf(stderr, "NOTE: too few datapoints in fit range to force the parabel falling -> taking tangente to get E\n");
	}
	k = -1 * intercept/slope;
	for(i=0; i<MAX_ALNS; i++) {
	  fit = L[i].alnscore*slope+intercept;
	  if(fit < DBL_MIN_EXP) L[i].evalue = 0;
	  else L[i].evalue = k * exp(fit);
	}
      }
	
      for(i=0; i<MAX_ALNS; i++) {
	if(L[i].alnscore == -1)break;
	if(arg.noE==0 && L[i].evalue >= 10*cfg.usrE) break;
	else if(cfg.percID != 0 && L[i].percID < cfg.percID) break;
 
	gotoh_printGotohAln(L[i], cfg.output_format, res_fp);

      }
      
    end0:

      gotoh_freeArrayD(x); gotoh_freeArrayD(y);
      
      if(cfg.res_file != NULL)
	fclose(res_fp);

      fflush(NULL);
    
      free(p);
      free(score_distr);
      free(q_seq);
      free(q_name);
      gotoh_freeArrayI(lmScores);

      for(i=0; i<MAX_ALNS; i++)
	if(L[i].alnscore > -1)
	  gotoh_freeL(L[i]);  
   
    }
    /* next query sequence -> */
    /* exit(0); */
  }
 
  CloseFASTA(fp_q);
  CloseFASTA(fp_db);
  
  return 0;
  
  /****************************************************************************/
  /* here starts the single motif calculation */
 motif1:

  mS = gotoh_readMotifSingle(arg.m1_file);

  struct farray scores_dbf;
  struct iarray lmScores2;
  
  /* gotoh_printMotifSingle(mS); */

  /* init a candidates array */
  for(i=0; i<MAX_ALNS; i++)
    L[i] = gotoh_initL();

  /* init a score array */
  lo = 2 * (cfg.gap_open + ((mS.len - 1) * cfg.gap_extend));
  up = cfg.match * mS.len;
  lo_transf = 0;           /* transformation: score += lo */
  up_transf = abs(up) + abs(lo); 
  /* size = up_transf + 1;  */   /* 1 => add 0 */
  size = abs(lo*10) + abs(up*10) + 1;

  /* holds the untransformed scores */
  lmScores2 = gotoh_initArrayI(size);

  /* distribution of all possible scores per query */
  score_distr = (int *) calloc(size + 1, sizeof(int));
  for(i=0; i<size; i++) {
    score_distr[i] = 0;
  }
  
  m = 2*mS.len;

  /* read and search database */
  fp_db = OpenFASTA(cfg.dbase); /* abfangen on error!!! */
  if(fp_db == NULL) {
    fprintf(stderr, "GS-ERROR: Cannot access database file: '%s'.\n",cfg.dbase);
    free(score_distr);
    gotoh_freeArrayI(lmScores2);
    exit(0);
  }
  
  while(ReadFASTA(fp_db, &db_seq, &db_name, &db_len)) {
  
    scores_dbf = gotoh_initArrayF(db_len+1); /*start with 1!*/

    char *rev_db_seq=NULL;
    rev_db_seq = gotoh_revCompl(db_seq, cfg.verbose, db_len);

    if(db_len < mS.len) {
      if(cfg.verbose == 1) {
	fprintf(stderr, "GS-WARNING: Database sequence shorter than motif.\n");
	fprintf(stderr, "        DB: %s length: %d\n",db_name,db_len);
	fprintf(stderr, "     QUERY: %s length: %d\n",mS.name,mS.len);
	fprintf(stderr, "...skipping %s.\n",db_name);
      }
      free(db_seq);
      free(db_name);

      continue;
	
    } 

    if(cfg.quiet == 0) { fprintf(stderr, "# Q: %s ==> DB: %s\n",mS.name,db_name); fflush(stderr); }
	
    /* gotoh_fwRecursionMotifSingle(&scores_dbf, db_seq, db_len, mS, db_name, score_distr, size, cfg, 0, '+'); */
    gotoh_fwRecursionMotifSingle(db_seq, db_len, mS, &scores_dbf, score_distr, cfg, '+');

    /* int o; for(o=0; o<scores_dbf.size; o++) printf("D %d %.2f\n",o,scores_dbf.vals[o]); exit(0); */
    
    if(cfg.quiet == 0) { fprintf(stderr, "\n#    Backtracing k best candidates "); fflush(stderr); }
    gotoh_findLocalMaximaMotifSingle(&scores_dbf, L, 0, db_len-1, db_seq, db_name, mS, '+', cfg, &lmScores2, db_len);
    
    /* int o,mod; */
    /* float sum=1; o = 0; */
    /* while(sum > 0) { */
    /*   sum = gotoh_getLocalMaxMotif(&scores_dbf, mS, L, cfg, db_seq, db_name, '+', &lmScores2); */
    /*   /\* mod = o%4; o++; *\/ */
    /*   /\* if(mod == 0) fprintf(stderr, "\r# |"); *\/ */
    /*   /\* else if(mod == 1) fprintf(stderr, "\r# /"); *\/ */
    /*   /\* else if(mod == 2) fprintf(stderr, "\r# -"); *\/ */
    /*   /\* else if(mod == 3) fprintf(stderr, "\r# \\"); *\/ */
    /* } */
    
    if(cfg.quiet == 0) { fprintf(stderr, "\r# .\n"); fflush(stderr); }
    
    /* for(i=0; i<scores_dbf.size; i++) { scores_dbf.vals[i] = 0.; } */

    gotoh_fwRecursionMotifSingle(rev_db_seq, db_len, mS, &scores_dbf, score_distr, cfg, '-');
    
    if(cfg.quiet == 0) { fprintf(stderr, "\n#    Backtracing k best candidates "); fflush(stderr); }
    gotoh_findLocalMaximaMotifSingle(&scores_dbf, L, 0, db_len-1, rev_db_seq, db_name, mS, '-', cfg, &lmScores2, db_len);
    /* sum=1; o = 0; */
    /* while(sum > 0) { */
    /*   sum = gotoh_getLocalMaxMotif(&scores_dbf, mS, L, cfg, rev_db_seq, db_name, '-', &lmScores2); */
    /*   /\* mod = o%4; o++; *\/ */
    /*   /\* if(mod == 0) fprintf(stderr, "\r# |"); *\/ */
    /*   /\* else if(mod == 1) fprintf(stderr, "\r# /"); *\/ */
    /*   /\* else if(mod == 2) fprintf(stderr, "\r# -"); *\/ */
    /*   /\* else if(mod == 3) fprintf(stderr, "\r# \\"); *\/ */
    /*   /\* printf("sum(-): %f\n",sum); *\/ */
    /* } */
    if(cfg.quiet == 0) { fprintf(stderr, "\r# .\n"); fflush(stderr); }
      
    N++; /* count database entries */
      
    gotoh_freeArrayF(scores_dbf);
      
    free(db_name);
    free(db_seq);
    free(rev_db_seq);

  }
  
  /* CloseFASTA(fp_db); */
  
  /* for(i=0; i<lmScores2.size; i++) fprintf(stdout, "LMH %d %d\n", i, lmScores2.vals[i]); */

  if(N > 0) {
    
    if(cfg.res_file != NULL)
      res_fp = fopen(cfg.res_file, "w");
    else
      res_fp = stdout;

    int k;
    struct darray x;
    struct darray y;
      
    x = gotoh_initArrayD(lmScores2.size);
    y = gotoh_initArrayD(lmScores2.size);

    k=0;
    for(i=0; i<lmScores2.size; i++) {
      if(lmScores2.vals[i] == 0) continue;
      x.vals[k] = (double)(i)/10; /* original score was *10 to not loose too many scores */
      y.vals[k] = (double)(log(lmScores2.vals[i]));
      /* printf("data %d %f\n", i, y.vals[k]); */
      k++;
    }

    /* printf("k: %d\n",k); */

    x.size = k;
    y.size = k;

    int ecalc=0;
    double slope, intercept;
      
    ecalc = gotoh_getFitRange(x, y, &startx, &cut, &slope, &intercept);

    /* printf("ecalc: %d\n", ecalc); */
    if(ecalc == 0) {
      if(cfg.quiet == 0) {
	fprintf(res_fp, "# Too few alignments to estimate evalue-calculation\n");
	fprintf(res_fp, "# Returning best canidate - which might not be a significant one!\n");
      }
      gotoh_printGotohAln(L[0], cfg.output_format, res_fp);
      goto end1;
    }
    /* printf("left: %d right: %d\n", startx, cut); */
  
    ecalc = gotoh_calcEvalue(L, MAX_ALNS, x, y, startx, cut, lmScores2.size);
    if(ecalc == 0) {
      /* too few datapoints in fit range to force the parabel falling */
      /* take tangente to get E */
      double k, fit;

      k = -1 * intercept/slope;
      for(i=0; i<MAX_ALNS; i++) {
	fit = L[i].alnscore*slope+intercept;
	if(fit < DBL_MIN_EXP) L[i].evalue = 0;
	else L[i].evalue = k * exp(fit);
      }
    }

    for(i=0; i<MAX_ALNS; i++) {
      if(L[i].alnscore == -1) break;
      if(arg.noE==0 && L[i].evalue >= 10* cfg.usrE) break;
      else if(cfg.percID != 0 && L[i].percID < cfg.percID) break;

      gotoh_printGotohAln(L[i], cfg.output_format, res_fp);

    }

  end1:
    gotoh_freeArrayD(x); gotoh_freeArrayD(y);
   
    if(cfg.res_file != NULL)
      fclose(res_fp);
     
  }
  
  for(i=0; i<MAX_ALNS; i++)
    if(L[i].alnscore > -1)
      gotoh_freeL(L[i]);  
  
  free(score_distr); gotoh_freeArrayI(lmScores2);
  gotoh_freeMotifSingle(mS);
  gotoh_freeArguments(arg);
    
  return 0;
    
  /****************************************************************************/
  /* here starts the 2-sided-motif calculation */
 motif2:
  
  gotoh_initBasePairTable();

  motif = gotoh_readMotif2sided(arg.m2_file, cfg);

  
  /* init a candidates array */
  for(i=0; i<MAX_ALNS; i++){
    L[i] = gotoh_initL();
    /* gotoh_printGotohAln(L[i]); */
  }
  
  /* init a score array */
  lo = 2 * (cfg.gap_open + ((motif.nLeft - 1) * cfg.gap_extend));
  up = cfg.match * motif.nLeft;
  lo_transf = 0;           /* transformation: score += lo */
  up_transf = abs(up) + abs(lo); 
  /* size = up_transf + 1;  */   /* 1 => add 0 */
  sizeL = abs(lo) + abs(up) + 1;

  lo = 2 * (cfg.gap_open + ((motif.nRight - 1) * cfg.gap_extend));
  up = cfg.match * motif.nRight;
  sizeR = abs(lo) + abs(up) + 1;

  /* distribution of all possible scores per query */
  score_histoL = (int *) calloc(sizeL + 1, sizeof(int));
  score_histoR = (int *) calloc(sizeR + 1, sizeof(int));
  for(i=0; i<sizeL; i++) score_histoL[i] = 0;
  for(i=0; i<sizeR; i++) score_histoR[i] = 0;

  m = 2*motif.len;
  
  fp_db = OpenFASTA(cfg.dbase); /* abfangen on error!!! */
  if(fp_db == NULL) {
    fprintf(stderr, "GS-ERROR: Cannot access database file: '%s'.\n",cfg.dbase);
    free(q_seq); free(q_name);
    free(score_histoL);
    free(score_histoR);
    exit(0);    
  }

  N = 0;
  while(ReadFASTA(fp_db, &db_seq, &db_name, &db_len)) {
    /* also scan revcompl of dbase! */
    /* initialize float score distr. to db_len -> for each db sequence extra! */
    score_distrL = (float *) calloc(db_len+1, sizeof(float));
    score_distrR = (float *) calloc(db_len+1, sizeof(float));
    for(i=0; i<db_len; i++) {
      score_distrL[i] = 0;
      score_distrR[i] = 0;
    }
    
    if(db_len < motif.len) {
      if(cfg.verbose == 1) {
	fprintf(stderr, "GS-WARNING: Database sequence shorter than query.\n");
	fprintf(stderr, "        DB: %s length: %d\n",db_name,db_len);
	fprintf(stderr, "     QUERY: %s length: %d\n",motif.name,motif.len);
	fprintf(stderr, "...skipping %s.\n",db_name);
      }
      free(db_seq);
      free(db_name);
      continue;
    }
    
    fprintf(stderr, "# M: %s ==> DB: %s\n",motif.name,db_name);
      
      
    gotoh_scanDbaseMotif2sided(db_seq, db_len, db_name, motif, score_histoL, score_histoR,
			       score_distrL, score_distrR, sizeL, sizeR, cfg, L, 0);

    fprintf(stderr, "#    100.00 %% (%d)\n", db_len);
    fflush(stderr);
      
    N++;
    /* do something with the float score distributions - db dependent! */
    /* for(i=0; i<db_len; i++) { */
    /*   printf("L  %i    %f\n",i,score_distrL[i]); */
    /*   printf("R  %i    %f\n",i,score_distrR[i]); */
    /* } */
    
  
    free(db_name);
    free(db_seq);
    free(score_distrL); free(score_distrR);
    

    /* move out of the loop when turning the loop on again */
    /* if(strchr(candy.left_seq, '-') != NULL || strchr(candy.right_seq, '-') != NULL || */
    /*    strchr(candy.left_mseq, '-') != NULL || strchr(candy.right_mseq, '-') != NULL) { */

      /* gotoh_freeMotif(motif); */
      
      /* gotoh_freeArguments(arg); */
      /* if(arg.cfg_file != NULL) gotoh_freeSettings(cfg); */
      /* CloseFASTA(fp_db); */
      /* /\* end moving .. *\/ */
      /* exit(0); */
    /* } */
    /* if(N == 1000) break; */
  }
  
  /* for(i=0; i<sizeL; i++) */
  /*   fprintf(stdout, "HL %i    %d\n",i, score_histoL[i]); */
  /* for(i=0; i<sizeR; i++) */
  /*   fprintf(stdout, "HR %i    %d\n",i, score_histoR[i]); */
  
  gotoh_freeMotif(motif);
  free(score_histoL); free(score_histoR);
  
  CloseFASTA(fp_db);

  /* here ends the motif calculation */
  /****************************************************************************/


  /* free all remaining stuff */
  gotoh_freeArguments(arg);
  if(arg.cfg_file != NULL) gotoh_freeSettings(cfg);
  
  return 0;
  
}
