#include "read.h"

char cur_q_seq[1000];
char next_q_seq[1000];
char cur_db_seq[1000];
char next_db_seq[1000];
extern float gotoh_bp[13][496];

/* -------------------------------------------------------------------------- */

char *get_line(FILE *fp) { /* reads lines of arbitrary length from fp */

   char s[512], *line, *cp;

   line = NULL;
   do {
      if (fgets(s, 512, fp)==NULL) break;
      cp = strchr(s, '\n');
      if (cp != NULL) *cp = '\0';
      if (line==NULL)
	line = (char *) calloc(strlen(s)+1, sizeof(char));
      else
         line = (char *) realloc(line, strlen(s)+strlen(line)+1);
      strcat(line,s);
   } while(cp==NULL);

   return line;
   
}

/* -------------------------------------------------------------------------- */

struct gotoh_arguments gotoh_readArguments(int argc, char *argv[]) {

  int i;
  gotoh_arguments args;

  if(argc < 2) 
    usage();
  
  args = gotoh_initArguments();
  
  for(i=1; i<argc; i++) {
    if(argv[i][0] == '-') {
      switch(argv[i][1]) {
      case 'h':
	usage();
	break;
      case 'v':
	version();
	break;
      case 'd':   /* database file */
	if(argc < i+2) usage();
	if(args.db_file == NULL) {
	  args.db_file = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	  strcpy(args.db_file,argv[i+1]);
	}
	break;
      case 'q':   /* query file */
	if(argc < i+2) usage();
	if(!args.q_file) {
	  args.q_file = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	  strcpy(args.q_file, argv[i+1]);
	}
        break;
      case 'm':   /* motif file */
	if(strcmp(argv[i], "-m1")==0) { /* single motif */
	  if(argc < i+2) usage();
	  if(args.m1_file == NULL) {
	    args.m1_file = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	    strcpy(args.m1_file, argv[i+1]);
	  }
	} else if(strcmp(argv[i], "-m2")==0) { /* two-sided motif */
	  if(argc < i+2) usage();
	  if(args.m2_file == NULL) {
	    args.m2_file = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	    strcpy(args.m2_file, argv[i+1]);
	  }
	}
	break;
      case 'c':   /* config file */
	if(argc < i+2) usage();
	if(args.cfg_file == NULL) {
	  args.cfg_file = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	  strcpy(args.cfg_file, argv[i+1]);
	}
	break;
      case 'e':   /* evalue */
	if(argc < i+2) usage();
	if(args.e == -1)   /* not set yet */
	  args.e = strtod(argv[i+1], NULL);
	break;
      case 'p':   /* percent identity */
	if(argc < i+2) usage();
	if(args.p < 0)   /* not set yet */
	  args.p = atof(argv[i+1]);
	break;
      case 'o':   /* output format */
	if(argc < i+2) usage();
	if(args.out == -1)   /* not set yet */
	  args.out = atoi(argv[i+1]);
	break;
      case 'r':   /* result file */
	if(argc < i+2) usage();
	if(args.res_file == NULL) {
	  args.res_file = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	  strcpy(args.res_file, argv[i+1]);
	}
	break;
      case '-':
	if(strcmp(argv[i],        "--help"     ) == 0) usage();
	else if(strcmp(argv[i],   "--dbase"    ) == 0) {
	  if(argc < i+2) usage();
	  if(args.db_file == NULL) {
	    args.db_file = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	    strcpy(args.db_file,argv[i+1]);
	  }
	} else if(strcmp(argv[i], "--query"    ) == 0) {
	  if(argc < i+2) usage();
	  if(args.q_file == NULL) {
	    args.q_file = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	    strcpy(args.q_file, argv[i+1]);
	  }
	} else if(strcmp(argv[i], "--motif1"   ) == 0) {
	  if(argc < i+2) usage();
	  if(args.m1_file == NULL) {
	    args.m1_file = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	    strcpy(args.m1_file, argv[i+1]);
	  }
	} else if(strcmp(argv[i], "--motif2"   ) == 0) {
	  if(argc < i+2) usage();
	  if(args.m2_file == NULL) {
	    args.m2_file = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	    strcpy(args.m2_file, argv[i+1]);
	  }
	} else if(strcmp(argv[i], "--config"   ) == 0) {
	  if(argc < i+2) usage();
	  if(args.cfg_file == NULL) {
	    args.cfg_file = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	    strcpy(args.cfg_file, argv[i+1]);
	  } 
	} else if(strcmp(argv[i], "--evalue"   ) == 0) {
	  if(argc < i+2) usage();
	  if(args.e == -1)  
	    args.e = strtod(argv[i+1], NULL);
	} else if(strcmp(argv[i], "--percent"  ) == 0) {
	  if(argc < i+2) usage();
	  if(args.p == 0) 
	    args.p = atof(argv[i+1]);
	} else if(strcmp(argv[i], "--m1Seq"  ) == 0) {
	  if(argc < i+2) usage();
	  if(args.m1_seqT == 0) 
	    args.m1_seqT = atof(argv[i+1]);
	} else if(strcmp(argv[i], "--m2Seq"  ) == 0) {
	  if(argc < i+2) usage();
	  if(args.m2_seqT == 0) 
	    args.m2_seqT = atof(argv[i+1]);
	} else if(strcmp(argv[i], "--m2Struc") == 0) {
	  if(argc < i+2) usage();
	  if(args.m2_strucT == 0) 
	    args.m2_strucT = atof(argv[i+1]);
	} else if(strcmp(argv[i], "--m2Dist")  == 0) {
	  if(argc < i+2) usage();
	  if(args.mD == 0) 
	    args.mD = atoi(argv[i+1]);
	} else if(strcmp(argv[i], "--output")  == 0) {
	  if(argc < i+2) usage();
	  if(args.out == -1)  
	    args.out = atoi(argv[i+1]);
	} else if(strcmp(argv[i], "--stats")   == 0) {
	  if(argc < i+2) usage();
	  if(args.stat_file == NULL) {
	    args.stat_file = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	    strcpy(args.stat_file, argv[i+1]);
	  }
	} else if(strcmp(argv[i], "--result"   ) == 0) {
	  if(argc < i+2) usage();
	  if(args.res_file == NULL) {
	    args.res_file = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	    strcpy(args.res_file, argv[i+1]);
	  }
	} else if(strcmp(argv[i], "--noE"       ) == 0) {
	  args.noE = 1;
	} else if(strcmp(argv[i], "--quiet"   ) == 0) {
	  args.quiet = 1;
	}
	break;
      default:
	usage();
      }
    } else if(argv[i][0] == '?') {
      usage();
    } 
  }


  /* check INPUT validity */
  if(args.cfg_file == NULL) {
    if(args.db_file == NULL) {
      gotoh_freeArguments(args);
      gotoh_Error("Specify database file OR  provide this information in a configuration file and embedd it with -c|--config.",args.verbose);
    }
    if(args.q_file == NULL && args.m1_file == NULL && args.m2_file == NULL) {
      gotoh_freeArguments(args);
      gotoh_Error("Specify query/motif file OR provide this information in a configuration file and embedd it with -c|--config.",args.verbose);
    }
  } else {
    if(args.db_file != NULL && (args.q_file != NULL || args.m1_file !=NULL || args.m2_file != NULL)) {
      gotoh_Warning("Input files as provided by configuration file are ignored, command line parameters are used.",args.verbose);
    } 
  }

  if(args.q_file != NULL && (args.m1_file != NULL || args.m2_file != NULL)) {
    gotoh_freeArguments(args);
    gotoh_Error("Specify EITHER query sequence OR motif file.",args.verbose);
  }

  return args;
  
}

/* -------------------------------------------------------------------------- */

struct gotoh_settings gotoh_readSettings(gotoh_arguments ga) {

  FILE *fp=NULL;
  char parameter[1024], value[1024], *line=NULL;
  gotoh_settings cfg;

  cfg = gotoh_initSettings();

  if(ga.cfg_file == NULL) {
    /* input */
    cfg.dbase         = ga.db_file;
    cfg.query         = ga.q_file;
    cfg.motif1        = ga.m1_file;
    cfg.motif2        = ga.m2_file;
    /* filter */
    cfg.usrE          = ga.e;
    cfg.percID        = ga.p;
    cfg.m1_seqT       = ga.m1_seqT;
    cfg.m2_seqT       = ga.m2_seqT;
    cfg.m2_strucT     = ga.m2_strucT;
    cfg.m2T           = (2*cfg.m2_seqT + cfg.m2_strucT) / 3;
    cfg.maxdistance   = ga.mD;
    /* output */
    cfg.output_format = ga.out;
    cfg.stat_file     = ga.stat_file;
    cfg.res_file      = ga.res_file;
    cfg.verbose       = ga.verbose;
    cfg.quiet         = ga.quiet;

    /* check VALIDITY of parameters */
    if(cfg.usrE == -1 || cfg.usrE <= 0 || cfg.usrE > 10) cfg.usrE = 1e-3;
    if(cfg.percID == 0 || cfg.percID < 0 || cfg.percID > 100) cfg.percID = 60;
    if(cfg.m1_seqT == -1 || cfg.m1_seqT < 0 || cfg.m1_seqT > 1) cfg.m1_seqT = 0.7;
    if(cfg.m2_seqT == -1 || cfg.m2_seqT < 0 || cfg.m2_seqT > 1) cfg.m2_seqT = 0.5;
    if(cfg.m2_strucT == -1 || cfg.m2_strucT < 0 || cfg.m2_strucT > 1) cfg.m2_strucT = 0.8;
    cfg.m2T           = (2*cfg.m2_seqT + cfg.m2_strucT) / 3;
    if(cfg.maxdistance == 0 || cfg.maxdistance < 0) cfg.maxdistance = 200;
    if(cfg.output_format < 0 || cfg.output_format > 6) cfg.output_format = 0;
    
    return cfg;
  }

  /* check VALIDITY of parameters */
  if(cfg.usrE == -1 || cfg.usrE < 0 || cfg.usrE > 10) cfg.usrE = 1e-3;
  if(cfg.percID == 0 || cfg.percID < 0 || cfg.percID > 100) cfg.percID = 60;
  if(cfg.m1_seqT == -1 || cfg.m1_seqT < 0 || cfg.m1_seqT > 1) cfg.m1_seqT = 0.7;
  if(cfg.m2_seqT == -1 || cfg.m2_seqT < 0 || cfg.m2_seqT > 1) cfg.m2_seqT = 0.5;
  if(cfg.m2_strucT == -1 || cfg.m2_strucT < 0 || cfg.m2_strucT > 1) cfg.m2_strucT = 0.8;
  cfg.m2T           = (2*cfg.m2_seqT + cfg.m2_strucT) / 3;
  if(cfg.maxdistance == 0 || cfg.maxdistance < 0) cfg.maxdistance = 200;
  if(cfg.output_format < 0 || cfg.output_format > 6) cfg.output_format = 0;
  
  fp = fopen(ga.cfg_file, "r");
  
  if(fp==NULL) {
    gotoh_freeArguments(ga);
    gotoh_Error("Error: can't access config file", ga.verbose);
  } else {
    while(!feof(fp)) {
      line = get_line(fp);
      /* line = fgets(s, 1024, fp); */
      if(line == NULL) break;
      
      if(line[0] == '#' || strlen(line) < 1) {
	free(line);
	continue;
      }
      sscanf(line, "%s %s", parameter, value);
      /* fscanf(fp, "%s %s", parameter, value); */
      /* fprintf(stderr, "%s %s\n",parameter, value); */
      if(line[0] == '#' || line[0] == '\0' || strlen(line) < 1) {
	continue;
      }
      if(strcmp(parameter,"DBASE")==0) {
	if(ga.db_file ==NULL) {
	  cfg.dbase = (char *) calloc(strlen(value)+1, sizeof(char));
	  strcpy(cfg.dbase, value);
	}
      } else if(strcmp(parameter, "QUERY") == 0) {
	if(ga.q_file == NULL) {
	  cfg.query = (char *) calloc(strlen(value)+1, sizeof(char));
	  strcpy(cfg.query, value);
	}
      } else if(strcmp(parameter, "MOTIF1") == 0) {
	if(ga.m1_file == NULL) {
	  cfg.motif1 = (char *) calloc(strlen(value)+1, sizeof(char));
	  strcpy(cfg.motif1, value);
	}
      } else if(strcmp(parameter, "MOTIF2") == 0) {
	if(ga.m2_file == NULL) {
	  cfg.motif2 = (char *) calloc(strlen(ga.m2_file)+1, sizeof(char));
	  strcpy(cfg.motif2, ga.m2_file);
	}
      } else if(strcmp(parameter, "GAPOPEN") == 0) {
	cfg.gap_open = atof(value);
      } else if(strcmp(parameter, "GAPEXTEND") == 0) {
	cfg.gap_extend = atof(value);
      } else if(strcmp(parameter, "MATCH") == 0) {
	cfg.match = atof(value);
      } else if(strcmp(parameter, "MISMATCH") == 0) {
	cfg.mismatch = atof(value);
      } else if(strcmp(parameter, "M1_SEQ_MATCH") == 0) {
	if(ga.m1_seqT != -1) cfg.m1_seqT = ga.m1_seqT;
	else cfg.m1_seqT = atof(value);
      } else if(strcmp(parameter, "M2_SEQ_MATCH") == 0) {
	if(ga.m2_seqT != -1) cfg.m2_seqT = ga.m2_seqT;
	else cfg.m2_seqT = atof(value);
      } else if(strcmp(parameter, "M2_STRUC_MATCH") == 0) {
	if(ga.m2_strucT != -1) cfg.m2_strucT = ga.m2_strucT;
	else cfg.m2_strucT = atof(value);
      } else if(strcmp(parameter, "M2_SUM_T") == 0) {
	if(ga.m2T != -1) cfg.m2T = ga.m2T;
	else cfg.m2T = atof(value);
      } else if(strcmp(parameter, "E-VALUE") == 0) { 
	if(ga.e != 999999) cfg.usrE = ga.e;
	else cfg.usrE = log10(strtod(value, NULL));
      } else if(strcmp(parameter, "PERCID") == 0) {
	if(ga.p != -1) cfg.percID = ga.p;
	else cfg.percID = atof(value);
      } else if(strcmp(parameter, "MAX_DISTANCE") == 0) {
	if(ga.mD != -1) cfg.maxdistance = ga.mD;
	else cfg.maxdistance = atoi(value);
      } else if(strcmp(parameter, "OUTPUT_FORMAT") == 0) {
	if(ga.out != -1) cfg.output_format = ga.out;
	else cfg.output_format = atof(value);
      } else if(strcmp(parameter, "RESULT_FILE") == 0) {
	if(cfg.res_file == NULL) {
	  cfg.res_file = (char *) calloc(strlen(value)+1, sizeof(char));
	  strcpy(cfg.res_file, value);
	} 
      } else if(strcmp(parameter, "STAT_FILE") == 0) {
	if(cfg.stat_file == NULL) {
	  cfg.stat_file = (char *) calloc(strlen(value)+1, sizeof(char));
	  strcpy(cfg.stat_file, value);
	}
      } else {
	gotoh_freeSettings(cfg);
	gotoh_Error("here is a problem with the format of the config file", cfg.verbose);
      }
      free(line);
    }
  }

  cfg.inf = 999999999;

  fclose(fp);

  return cfg;
  
}

/* -------------------------------------------------------------------------- */

struct gotoh_motifSingle gotoh_readMotifSingle(char *file) {

  FILE *fp = NULL;
  char *line=NULL;
  int maxLineLength = 10000;
  struct gotoh_motifSingle mS;
  char *token, separator[]=" ";
  int i,j;
  float min, max, sum;
  
  mS.name = (char *) calloc(strlen(file)+1, sizeof(char));
  strcpy(mS.name, file);
  
  fp = fopen(file, "rd");

  /* get consensus sequence */
  line = (char *) calloc(maxLineLength, sizeof(char));

  fgets(line, maxLineLength, fp);
  mS.maxC = (char *) calloc(strlen(line)+1, sizeof(char));
  token = strtok(line, separator);
  if(token != NULL) token = strtok(NULL, separator);
  
  while(token != NULL) {
    strcat(mS.maxC, token);
    token = strtok(NULL, separator);
  }
  free(line);

  j = strlen(mS.maxC);
  mS.maxC[j-1] = '\0';

  mS.len = strlen(mS.maxC);
  /* printf("ms.len: %d\n%s\n",mS.len,mS.maxC); */

  /* get frequencies */
  mS.freq = (float **) calloc(5, sizeof(float *));
  for(i=0; i<5; i++) 
    mS.freq[i] = (float *) calloc(mS.len+1, sizeof(float));
  
  for(i=0; i<5; i++) {
    line = (char *) calloc(maxLineLength, sizeof(char));
    fgets(line, maxLineLength, fp);
    /* parse line splitted by spaces.. */
    token = strtok(line, separator);        /* erstes Token ermitteln */
    /* mS.freq[i][0] = atof(token); */
    j=0;
    while( token != NULL && j<mS.len ) {
      mS.freq[i][j] = atof(token); j++;
      token = strtok(NULL, separator);        /* weitere Token ermitteln */
    }
    free(line);
  }
    
  fclose(fp);

  /* calc min, max, info */
  mS.min  = (float *) calloc(mS.len + 1, sizeof(float));
  mS.max  = (float *) calloc(mS.len + 1, sizeof(float));
  mS.info = (float *) calloc(mS.len + 1, sizeof(float));

  sum = 0;
  for(j=0; j<mS.len; j++) {
    min = mS.freq[0][j];
    max = mS.freq[0][j];
    sum = 0;
    for(i=1; i<5; i++) {
      if(mS.freq[i][j] < min) min = mS.freq[i][j];
      if(mS.freq[i][j] > max)
	max = mS.freq[i][j];

      if(mS.freq[i][j] > 0)
	sum += mS.freq[i][j] * log2(mS.freq[i][j]);
    }
    mS.min[j] = min;
    mS.max[j] = max;
     
    mS.info[j] = 2 + sum;
    
  }

  return mS;
  
}

/* -------------------------------------------------------------------------- */

struct gotoh_motif gotoh_readMotif2sided(char *file, gotoh_settings cfg) {

  FILE *fp=NULL;
  char *line=NULL;
  struct gotoh_motif m;
  char *token, separator[]=" ";
  int i,j, s, q, t;
  float min, max, sum, sumMaxL, sumMaxR;
  char bases2[] = "ACGU";
  
  fp = fopen(file, "rd");

  line = (char *) calloc(1024, sizeof(char));
  
  fgets(line, 1024, fp);
  m.comment = (char *) calloc(strlen(line)+1, sizeof(char));
  strcpy(m.comment, strtok(line, "\n"));

  free(line);

  /* get name from the first comment word */
  m.name = (char *) calloc(1024, sizeof(char));
  /* strcpy(m.name, m.comment); */
  m.comment += 2;
  strcpy(m.name, m.comment);
  m.comment -= 2;
  /* get left stem */
  line = (char *) calloc(1024, sizeof(char));
  fgets(line, 1024, fp);
  sscanf(line, "%d", &m.nLeft);
  free(line);

  m.stemL = (float **) calloc(5, sizeof(float *));
  for(i=0; i<5; i++)
    m.stemL[i] = (float *) calloc(m.nLeft + 1, sizeof(float));

  for(i=0; i<5; i++) {
    line = (char *) calloc(1024, sizeof(char));
    fgets(line, 1024, fp);
    /* parse line splitted by spaces.. */
    token = strtok(line, separator);        /* erstes Token ermitteln */
    /* m.stemL[i][0] = atof(token); */
    j=0;
    while( token != NULL && j<= m.nLeft) {
      m.stemL[i][j] = atof(token); j++;
      token = strtok(NULL, separator);        /* weitere Token ermitteln */
   }
    free(line);
  }
  
  line = (char *) calloc(1024, sizeof(char));
  while(strlen(line) == 0 || strncmp(line,"#",1) == 0) { /* ignore consensus sequence */
    free(line);
    line = (char *) calloc(1024, sizeof(char));
    fgets(line, 1024, fp);
  }
  
  sscanf(line, "%d ", &m.nRight);
  free(line);

  m.stemR = (float **) calloc(5, sizeof(float *));
  for(i=0; i<5; i++)
    m.stemR[i] = (float *) calloc(m.nRight + 1, sizeof(float));

  for(i=0; i<5; i++) {
    line = (char *) calloc(1024, sizeof(char));
    fgets(line, 1024, fp);
    /* parse line splitted by spaces.. */
    token = strtok(line, separator);        /* erstes Token ermitteln */
    j=0;
    while( token != NULL && j<= m.nRight ) {
      m.stemR[i][j] = atof(token);
      j++;
      token = strtok(NULL, separator);        /* weitere Token ermitteln */
    }
    free(line);
  }

  m.basepairs = (int **) calloc(m.nLeft + 1, sizeof(int *));
  for(i=0; i<m.nLeft; i++) {
    m.basepairs[i] = (int *) calloc(m.nRight + 1, sizeof(int));
    for(j=0; j<m.nRight; j++) 
      m.basepairs[i][j] = 0;
  }

  line = (char *) calloc(1024, sizeof(char));
  while(fgets(line, 1024, fp)) {

    token = strtok(line, separator);
    i = atoi(token);

    token = strtok(NULL, separator);
    j = atoi(token);

    token = strtok(NULL, separator);
    m.basepairs[i][j] = atoi(token);
    
    free(line);
    line = (char *) calloc(1024, sizeof(char));
  }

  free(line);
  fclose(fp);

  /* min length of whole motif */
  m.len = m.nLeft + m.nRight + 3;
  
  /* get minL, maxL and infoL */
  m.minL  = (float *) calloc(m.nLeft + 1, sizeof(float));
  m.maxL  = (float *) calloc(m.nLeft + 1, sizeof(float));
  m.infoL = (float *) calloc(m.nLeft + 1, sizeof(float));
  m.maxCL = (int *)   calloc(m.nLeft + 1, sizeof(int));
  m.consL = (char *)  calloc(m.nLeft + 1, sizeof(char));
  
  sum = 0; sumMaxL = 0; sumMaxR = 0;
  for(j=0; j<m.nLeft; j++) {
    min = m.stemL[0][j];
    max = m.stemL[0][j];
    sum = 0;
    for(i=1; i<4; i++) { /* ignore gaps */
      if(m.stemL[i][j] < min) min = m.stemL[i][j];
      if(m.stemL[i][j] > max) {
	max = m.stemL[i][j];
	m.maxCL[j] = i;  /* log if maxscored val is gap ! */
      }
      if(m.stemL[i][j] > 0)
	sum += m.stemL[i][j] * log2(m.stemL[i][j]);
    }
    m.minL[j] = min;
    m.maxL[j] = max;

    m.infoL[j] = 2 + sum;

    m.consL[j] = bases2[m.maxCL[j]];

    sumMaxL += max;
  }

  
  m.minR  = (float *) calloc(m.nRight + 1, sizeof(float));
  m.maxR  = (float *) calloc(m.nRight + 1, sizeof(float));
  m.infoR = (float *) calloc(m.nRight + 1, sizeof(float));
  m.maxCR = (int *)   calloc(m.nRight + 1, sizeof(int));
  m.consR = (char *)  calloc(m.nRight + 1, sizeof(char));

  sum = 0;
  for(j=0; j<m.nRight; j++) {
    min = m.stemR[0][j];
    max = m.stemR[0][j];
    sum = 0;
    for(i=1; i<5; i++) { /* do not ignore gaps */
      if(m.stemR[i][j] < min) min = m.stemR[i][j];
      if(m.stemR[i][j] > max) {
	max = m.stemR[i][j];
	m.maxCR[j] = i;
      }
      if(m.stemR[i][j] > 0) 
	sum += m.stemR[i][j] * log2(m.stemR[i][j]);
    }
    m.minR[j] = min;
    m.maxR[j] = max;

    m.infoR[j] = 2 + sum;

    m.consR[j] = bases2[m.maxCR[j]];

    sumMaxR += max;
    
  }

  /* get bp score for consensus */
  m.scoreBP = 0; m.n_bp = 0;
  for(i=0; i<m.nLeft; i++) {
    for(j=0; j<m.nRight; j++) {
      t = m.basepairs[i][j];
      if(t > 0) {
	s = m.consL[i];
	switch(s) {
	case 65:
	  q = m.consR[j];
	  break;
	case 67:
	  q = m.consR[j] + 128;
	  break;
	case 71:
	  q = m.consR[j] + 256;
	  break;
	case 85:
	  q = m.consR[j] + 384;
	  break;
	default:
	  q = m.consR[j];
	}
	m.scoreBP += gotoh_bp[t][q];
	m.n_bp++;
      }
    }
  }
  m.scoreBP /= m.n_bp;
    
  m.maxScore = m.nLeft*cfg.match + m.nRight*cfg.match + m.scoreBP;
 
  return m;
  
}

/* -------------------------------------------------------------------------- */

struct gotoh_arguments gotoh_initArguments() {

  gotoh_arguments a;

  a.cfg_file = NULL;
  a.db_file = NULL;
  a.q_file = NULL;
  a.res_file = NULL;
  a.stat_file = NULL;
  a.m1_file = NULL;
  a.m2_file = NULL;
  a.mD = 0;
  a.e = -1;
  a.p = -1;
  a.out = -1;
  a.help = 0;
  a.version = 0;
  a.verbose = 0;
  a.quiet = 0;
  a.m1_seqT = -1;
  a.m2_seqT = -1;
  a.m2_strucT = -1;
  a.m2T = (2*a.m2_seqT + a.m2_strucT) / 3;
  a.noE = 0;
  
  return a;
  
}

/* -------------------------------------------------------------------------- */

struct gotoh_settings gotoh_initSettings() {

  gotoh_settings s;

  /* input */
  s.dbase=NULL;
  s.query=NULL;
  s.motif1=NULL;
  s.motif2=NULL;
  s.res_file = NULL;
  /* alignment parameters / scores */
  s.gap_open = -8;
  s.gap_extend = -2;
  s.match = 3;
  s.mismatch = -1;
  s.inf = 99;
  /* filter */
  s.maxdistance = 200;
  s.usrE = 1e-3;
  s.percID = 60;
  s.m1_seqT = 0.7;
  s.m2_seqT = 0.5;
  s.m2_strucT = 0.8;
  s.m2T = (2*s.m2_seqT + s.m2_strucT) / 3;
  /* output */
  s.output_format = 0;
  s.stat_file=NULL;
  s.verbose = 0;
  s.quiet = 0;
  
  return s;
  
}

/* print functions ---------------------------------------------------------- */

void gotoh_printArguments(gotoh_arguments ga) {

  fprintf(stdout, "\nArguments:\n----------\n");
  fprintf(stdout, "Config file: %s\n",ga.cfg_file);
  fprintf(stdout, "Dbase file:  %s\n",ga.db_file);
  if(ga.res_file != NULL) fprintf(stdout, "Result file: %s\n",ga.res_file);
  if(ga.q_file != NULL) fprintf(stdout, "Query file:  %s\n",ga.q_file);
  if(ga.m1_file != NULL) {
    fprintf(stdout, "Motif file:  %s\n",ga.m1_file);
    fprintf(stdout, "M1_seq threshold: %f\n",ga.m1_seqT);
  }
  if(ga.m2_file != NULL)  {
    fprintf(stdout, "Motif file:         %s\n",ga.m2_file);
    fprintf(stdout, "M2_seq threshold:   %f\n",ga.m2_seqT);
    fprintf(stdout, "M2_struc threshold: %f\n",ga.m2_strucT);
    fprintf(stdout, "M2 threshold:       %f\n",ga.m2T);
  }
  
  fprintf(stdout, "Help:        %d\n",ga.help);
  fprintf(stdout, "Version:     %d\n",ga.version);

}

/* -------------------------------------------------------------------------- */

void gotoh_printSettings(gotoh_settings gs) {

  fprintf(stdout, "\nSettings:\n---------\n");
  fprintf(stdout, "Dbase file: %s\n",gs.dbase);
  if(gs.query != NULL) fprintf(stdout, "Query file: %s\n",gs.query);
  if(gs.res_file != NULL) fprintf(stdout, "Result file: %s\n",gs.res_file);
  if(gs.motif1 != NULL) {
    fprintf(stdout, "Motif file: %s\n",gs.motif1);
    fprintf(stdout, "M1_seq threshold: %f\n",gs.m1_seqT);
  }
  if(gs.motif2 != NULL) {
    fprintf(stdout, "Motif file: %s\n",gs.motif2);
    fprintf(stdout, "M2_seq threshold: %f\n",gs.m2_seqT);
    fprintf(stdout, "M2_struc threshold: %f\n",gs.m2_strucT);
    fprintf(stdout, "M2 threshold:       %f\n",gs.m2T);
  }
  
  fprintf(stdout, "Gap open:   %3d\n", gs.gap_open);
  fprintf(stdout, "Gap extend: %3d\n", gs.gap_extend);
  fprintf(stdout, "Match:      %3d\n", gs.match);
  fprintf(stdout, "Mismatch:   %3d\n", gs.mismatch);
  fprintf(stdout, "Inf:        %d\n", gs.inf);

  fprintf(stdout, "E-value:    %.2e\n", pow(10,gs.usrE));
  fprintf(stdout, "Perc Ident.:%5.2f\n", gs.percID);
  fprintf(stdout, "Output:     %3d\n", gs.output_format);

}

/* -------------------------------------------------------------------------- */

void gotoh_printMotifSingle(struct gotoh_motifSingle m) {

  int i,j;

  fprintf(stdout, "\nMotif:\n---------\n");
  fprintf(stdout, "Name:    %s\n", m.name);
  fprintf(stdout, "Length:  %d\n", m.len);

  for(i=0; i<5; i++) {
    for(j=0; j<m.len; j++) {
      fprintf(stdout, "%f   ", m.freq[i][j]);
    }
    fprintf(stdout, "\n");
  }
  fprintf(stdout, "\n");
  for(j=0; j<m.len; j++) 
    fprintf(stdout, "%f   ", m.min[j]);
  fprintf(stdout, "min\n");
  for(j=0; j<m.len; j++) 
    fprintf(stdout, "%f   ", m.max[j]);
  fprintf(stdout, "max\n");
  for(j=0; j<m.len; j++) 
    fprintf(stdout, "%f   ", m.info[j]);
  fprintf(stdout, "info\n");
  for(j=0; j<m.len; j++) 
    fprintf(stdout, "    %c      ", m.maxC[j]);
  fprintf(stdout, "consensus\n");
  
  
}

/* -------------------------------------------------------------------------- */

void gotoh_printMotif(gotoh_motif m) {

  int i, j;

    fprintf(stdout, "\nMotif:\n---------\n");
    fprintf(stdout, "Comment: %s\n", m.comment);
    fprintf(stdout, "Name:    %s\n", m.name);
    fprintf(stdout, "Length:  %d\n", m.len);
    fprintf(stdout, "nLeft:   %d =>", m.nLeft);
  
    fprintf(stdout, "left stem:\n");
    for(i=0; i<5; i++) {
      for(j=0; j<m.nLeft; j++) {
	fprintf(stdout, "%f\t", m.stemL[i][j]);
      }
      fprintf(stdout, "\n");
    }
    fprintf(stdout, "\n");
    for(j=0; j<m.nLeft; j++) 
      fprintf(stdout, "%f\t", m.minL[j]);
    fprintf(stdout, "min\n");
    for(j=0; j<m.nLeft; j++) 
      fprintf(stdout, "%f\t", m.maxL[j]);
    fprintf(stdout, "max\n");
    for(j=0; j<m.nLeft; j++) 
      fprintf(stdout, "%f\t", m.infoL[j]);
    fprintf(stdout, "info\n");
    for(j=0; j<m.nLeft; j++) 
      fprintf(stdout, "%8d\t", m.maxCL[j]);
    fprintf(stdout, "max symbol\n");
    for(j=0; j<m.nLeft; j++) 
      fprintf(stdout, "%8c\t", m.consL[j]);
    fprintf(stdout, "consensus\n");
  
    fprintf(stdout, "\nnRight:  %d =>", m.nRight);
    fprintf(stdout, "right stem:\n");
    for(i=0; i<4; i++) {
      for(j=0; j<m.nRight; j++) {
	fprintf(stdout, "%f\t", m.stemR[i][j]);
      }
      fprintf(stdout, "\n");
    }
    fprintf(stdout, "\n");
    for(j=0; j<m.nRight; j++) 
      fprintf(stdout, "%f\t", m.minR[j]);
    fprintf(stdout, "min\n");
    for(j=0; j<m.nRight; j++) 
      fprintf(stdout, "%f\t", m.maxR[j]);
    fprintf(stdout, "max\n");
    for(j=0; j<m.nRight; j++) 
      fprintf(stdout, "%f\t", m.infoR[j]);
    fprintf(stdout, "info\n");
    for(j=0; j<m.nRight; j++) 
      fprintf(stdout, "%8d\t", m.maxCR[j]);
    fprintf(stdout, "max symbol\n");
    for(j=0; j<m.nRight; j++) 
      fprintf(stdout, "%8c\t", m.consR[j]);
    fprintf(stdout, "consensus\n");

    fprintf(stdout, "\nBase pair table:\n");
    for(i=0; i<m.nLeft; i++) {
      for(j=0; j<m.nRight; j++)
	fprintf(stdout, "%5d", m.basepairs[i][j]);
      fprintf(stdout, "\n");
    }

    fprintf(stdout, "Number of base pairs: %d\n", m.n_bp);
    fprintf(stdout, "BP score of consensus: %f\n", m.scoreBP);
    fprintf(stdout, "\nScore of consensus: %f\n",m.maxScore);

}

/* -------------------------------------------------------------------------- */

void usage() {

  fprintf(stdout, "\n=============         GotohScan 2.0         =============\n\n");

  fprintf(stdout, "Call it: GotohScan arguments\n\n");
  fprintf(stdout, "The following arguments are possible (some required):\n\n");
  fprintf(stdout, "INPUT:\n\n");
  fprintf(stdout, "   %-20s Input database FILE in FASTA format. [REQUIRED]\n", "-d,--dbase FILE");
  fprintf(stdout, "   %-20s Input query FILE in FASTA format.\n", "-q,--query FILE");
  fprintf(stdout, "   %-20s Input single motif FILE\n", "-m1,--motif1 FILE");
  fprintf(stdout, "   %-20s formatted with scripts/fastaALN2frequency.pl\n", "");
  fprintf(stdout, "   %-20s Input motif FILE for 2-sided motifs like K-turns, etc.\n", "-m2,--motif2 FILE");
  fprintf(stdout, "   %-20s Format see examples/Kt11_web.motif.\n", "");
  fprintf(stdout, "   %-20s Input configuration FILE. \n", "-c,--config FILE");
  fprintf(stdout, "   %-20s See examples/Configuration.dat for an example.\n", "");
  fprintf(stdout, "   %-20s Note: Command line arguments overwrite configuration file!\n", "");
  fprintf(stdout, "\n   NOTE: Either a sequence (-q) or a motif (-m1/-m2) is REQUIRED as input!\n");

  fprintf(stdout, "\nSETTINGS:\n\n");
  fprintf(stdout, "   %-20s Database is splitted into NUMBER nt large subsequences.\n", "--split NUMBER");
  fprintf(stdout, "   %-20s Default: 10000\n", "");
  fprintf(stdout, "   %-20s Don't print status output of current query.\n", "--quiet");
  fprintf(stdout, "\n   Values for match, mismatch, gap-open and gap-extend can only be set in the\n");
  fprintf(stdout, "   configuration file. Default values are 3, -1, -8 and -2, resp.\n");
  
  
  fprintf(stdout, "\nFILTER:\n\n");
  fprintf(stdout, "%-20s Set Evalue (double!). NUMBER should be in [0, 10]. Default: 1e-3\n","-e NUMBER");
  fprintf(stdout, "%-20s Set percent identity of aligned sequences.\n","-p NUMBER");
  fprintf(stdout, "%-20s NUMBER should be in [0.0,100.0]\n","");
  fprintf(stdout, "%-20s Set threshold for a motif to be regarded as match.\n","-t NUMBER");
  fprintf(stdout, "%-20s Should be in [0,1]. Default: 0.5\n","");
  fprintf(stdout, "%-20s Print score distribution data for each query to a file. Default: unset\n","-s");
  fprintf(stdout, "%-20s Produces an xmgrace (.agr) file!\n","");
  fprintf(stdout, "%-20s Give output format. Default: 0\n", "-o NUMBER");
  fprintf(stdout, "%-20s \t0 - Blast tabular output\n","");
  fprintf(stdout, "%-20s \t1 - Blast tabular output + aligned sequences\n","");
  fprintf(stdout, "%-20s \t2 - FASTA format. NOTE: Hit sequence only, without gaps !\n","");
  fprintf(stdout, "%-20s \t3 - MAF format. NOTE: Header truncated to 30 characters!\n","");
  fprintf(stdout, "%-20s \t4 - BED + aligned sequences\n","");
  fprintf(stdout, "%-20s \t5 - GFF + aligned sequences\n\n","");
  /* fprintf(stdout, "%-20s Print Warnings and Notes. Default: 0\n\n","--verbose 0|1"); */
  fprintf(stdout, "%-20s Show this help message.\n", "-h,--help");
  fprintf(stdout, "%-20s Show version information.\n\n", "-v,--version");
  fprintf(stdout, "\nAlignment parameters and all other options can\n");
  fprintf(stdout, "also be set in a configuration file\n");
  fprintf(stdout, "see 'settings.cfg' as an example.\n\n");
  fprintf(stdout, "\nPlease feel free to contact me for comments, bug-reports, etc.\n\n");

  version();

}

/* -------------------------------------------------------------------------- */

void version() {
  
  fprintf(stdout, "\n=============         GotohScan 2.0         =============\n\n");

  fprintf(stdout, "Auhthor: Jana Hertel:\n");
  fprintf(stdout, "         jana@bioinf.uni-leipzig.de\n\n");
  fprintf(stdout, "Date:    October, 2010\n\n");
  
  exit(EXIT_SUCCESS);
  
}

/* free functions ----------------------------------------------------------- */

void gotoh_freeArguments(gotoh_arguments ga) {

  if(ga.cfg_file != NULL) free(ga.cfg_file);
  if(ga.db_file != NULL) free(ga.db_file);
  if(ga.q_file != NULL) free(ga.q_file);
  if(ga.res_file != NULL) free(ga.res_file);
  if(ga.m1_file != NULL) free(ga.m1_file);
  if(ga.m2_file != NULL) free(ga.m2_file);
  
}

/* -------------------------------------------------------------------------- */

void gotoh_freeSettings(gotoh_settings gs) {

  if(gs.dbase != NULL) free(gs.dbase);
  if(gs.query != NULL) free(gs.query);
  if(gs.res_file != NULL) free(gs.res_file);
  if(gs.motif1 != NULL) free(gs.motif1);
  if(gs.motif2 != NULL) free(gs.motif2);
  
}

/* -------------------------------------------------------------------------- */

void gotoh_freeMotifSingle(gotoh_motifSingle m) {

  int i;

  if(m.name) free(m.name);
  if(m.freq) {
    for(i=0; i<5; i++)
      if(m.freq[i]) free(m.freq[i]);
    free(m.freq);
  }
  if(m.min) free(m.min);
  if(m.max) free(m.max);
  if(m.info) free(m.info);
  if(m.maxC) free(m.maxC);

}

/* -------------------------------------------------------------------------- */

void gotoh_freeMotif(gotoh_motif gm) {

  int i;
  
  if(gm.comment != NULL) free(gm.comment);

  if(gm.name != NULL) free(gm.name);
  
  if(gm.stemL !=NULL) 
    for(i=0; i<5; i++)
      if(gm.stemL[i] != NULL) free(gm.stemL[i]);
  free(gm.stemL);

  if(gm.stemR !=NULL) 
    for(i=0; i<5; i++)
      if(gm.stemR[i] != NULL) free(gm.stemR[i]);
  free(gm.stemR);

  if(gm.minL != NULL)  free(gm.minL);
  if(gm.maxL != NULL)  free(gm.maxL);
  if(gm.infoL != NULL) free(gm.infoL);
  if(gm.maxCL != NULL) free(gm.maxCL);
  if(gm.consL != NULL) free(gm.consL);
  
  if(gm.minR != NULL)  free(gm.minR);
  if(gm.maxR != NULL)  free(gm.maxR);
  if(gm.infoR != NULL) free(gm.infoR);
  if(gm.maxCR != NULL) free(gm.maxCR);
  if(gm.consR != NULL) free(gm.consR);
  
  if(gm.basepairs) 
    for(i=0; i<gm.nLeft; i++)
      if(gm.basepairs[i])
	free(gm.basepairs[i]);
  free(gm.basepairs);
  
}


/* other functions ---------------------------------------------------------- */

void gotoh_Error(char *msg, int v) {

  if(v == 1) {
    fprintf(stderr, "GS-ERROR: %s\n",msg);
    exit(EXIT_FAILURE);
  }
}

/* -------------------------------------------------------------------------- */

void gotoh_Warning(char *msg, int v) {

  if(v == 1)
    fprintf(stderr, "GS-WARNING: %s\n",msg);
  
}

/* -------------------------------------------------------------------------- */


