/* singleSequenceCALC.c */
#include "singleSequenceCALC.h"

extern char bases[]; /* ACGUTRYMKWSBDHVNacgutrymkwsbdhvnt */
extern int gotoh_Table[128][128];

/* int nalns = 0; */

/* -------------------------------------------------------------------------- */

void gotoh_initMatchTable(gotoh_settings gs) {

  int i, j, ni, nj, score;

  for(i=0; i<128; i++) {
    for(j=0; j<128; j++) {
      if(i>j) continue;
      gotoh_Table[j][i] = gs.mismatch;
      gotoh_Table[i][j] = gs.mismatch;
    }
  }

  for(i=0; i<16; i++) {
    ni = bases[i]; /* i pos in table */
    for(j=0; j<16; j++) {
      if(i>j) continue;
      nj = bases[j]; /* j pos in table */
      if(i < 5 && i == j) {
	gotoh_Table[ni][nj] = gs.match;
	gotoh_Table[ni+32][nj] = gs.match;
	gotoh_Table[ni][nj+32] = gs.match;
	gotoh_Table[nj][ni] = gs.match;
	gotoh_Table[nj+32][ni] = gs.match;
	gotoh_Table[nj][ni+32] = gs.match;
	gotoh_Table[ni+32][nj+32] = gs.match;
	gotoh_Table[nj+32][ni+32] = gs.match;
      } else if(i < 11 && i == j) {
	score = floor(0.5 * gs.match + 0.5 * gs.mismatch);
	gotoh_Table[ni][nj] = score;
	gotoh_Table[ni+32][nj] = score;
	gotoh_Table[ni][nj+32] = score;
	gotoh_Table[nj][ni] = score;
	gotoh_Table[nj+32][ni] = score;
	gotoh_Table[nj][ni+32] = score;
	gotoh_Table[ni+32][nj+32] = score;
	gotoh_Table[nj+32][ni+32] = score;
      } else if(i < 15 && i == j) {
	score = floor((float) gs.match / 3 + (float)gs.mismatch / 3);
	gotoh_Table[ni][nj] = score;
	gotoh_Table[ni+32][nj] = score;
	gotoh_Table[ni][nj+32] = score;
	gotoh_Table[nj][ni] = score;
	gotoh_Table[nj+32][ni] = score;
	gotoh_Table[nj][ni+32] = score;
	gotoh_Table[ni+32][nj+32] = score;
	gotoh_Table[nj+32][ni+32] = score;
      } else if(i == 16 && j == 16) {
	score = floor((float) gs.match / 4 + (float)gs.mismatch / 4);
	gotoh_Table[ni][nj] = score;
	gotoh_Table[ni+32][nj] = score;
	gotoh_Table[ni][nj+32] = score;
	gotoh_Table[nj][ni] = score;
	gotoh_Table[nj+32][ni] = score;
	gotoh_Table[nj][ni+32] = score;
	gotoh_Table[ni+32][nj+32] = score;
	gotoh_Table[nj+32][ni+32] = score;
      }
    }
  }
  
  /* [U,T] -> [u,t] = match ! */
  gotoh_Table[84][85] = gs.match;
  gotoh_Table[84][117] = gs.match;
  gotoh_Table[85][84] = gs.match;
  gotoh_Table[85][116] = gs.match;
  gotoh_Table[116][85] = gs.match;
  gotoh_Table[117][84] = gs.match;
  gotoh_Table[116][117] = gs.match;
  gotoh_Table[117][116] = gs.match;
  /* [u,t] -> [U,T]  = match ! */
  
  return;

}

/* -------------------------------------------------------------------------- */

void gotoh_fwRecursion(char *s, int n, char *q, int m, struct iarray *scores, int *score_distr,
		       struct gotoh_settings cfg, char strand) {

  int  i,j,p,pp;
  int *S, *S_, *D, *D_, *F, *F_, *tmp, *G, *G_;
  int  transf_score, minAS;

  S   = (int *) calloc(m+1, sizeof(int));
  S_  = (int *) calloc(m+1, sizeof(int));
  D   = (int *) calloc(m+1, sizeof(int));
  D_  = (int *) calloc(m+1, sizeof(int));
  F   = (int *) calloc(m+1, sizeof(int));
  F_  = (int *) calloc(m+1, sizeof(int));
  G   = (int *) calloc(m+1, sizeof(int)); /* save number of gaps */
  G_  = (int *) calloc(m+1, sizeof(int));

  minAS = abs( 2 * (cfg.gap_open + ((m - 1) * cfg.gap_extend)));
  
  /* initialize semi-local alignment */
  D_[0] = -(cfg.inf);
  F_[0] = -(cfg.inf);
  S_[0] = 0;
  G_[0] = 0;

  for(j=1;j<=m;j++) { 
    D_[j] = -(cfg.inf);
    F_[j] = (j-1)*cfg.gap_extend + cfg.gap_open;
    S_[j] = F[j];
    G_[j] = j;
  }

  for(i=1;i<=n;i++) {
    if(i%1000==0 && cfg.quiet == 0) {
      fprintf(stderr, "\r#    %-5.2f %% (db length: %d) %c strand", (float) (i)/ (float) (n) * 100, n, strand);
      fflush(stderr);
    }
    /* init */
    D[0] = 0;  /* local alignment!*/
    F[0] = -cfg.inf;
    S[0] = 0;
    G[0] = 0;

    /* align query  against each database position i */
    for(j=1;j<=m;j++) {
      /* int p,pp; */
      p  = S[j-1] + cfg.gap_open;
      pp = D[j-1] + cfg.gap_extend;
      if(p>pp) D[j] = p;   /* start new insertion */
      else D[j] = pp;      /* extend insertion    */
      
      p = S_[j]  + cfg.gap_open;
      pp = F_[j]  + cfg.gap_extend;
      if(p>pp) F[j] = p;   /* start new deletion */
      else F[j] = pp;      /* extend deletion    */
      
      p = S_[j-1];

      /* get match/mismatch score from lookuptable */
      p += gotoh_Table[(unsigned char)s[i-1]][(unsigned char)q[j-1]];
      
      /* count gaps to get start position of alignment */

      pp = maxI(maxI(F[j],D[j]),p);

      if(pp == p) { G[j] = G_[j-1]; }
      else if(pp == F[j]) { G[j] = G_[j] + 1;}
      else if(pp == D[j]) { G[j] = G[j-1] + 1; }

      S[j] = pp; continue;
      
      /* older version */
      if(D[j]>p) p=D[j];                   /* insertion */
      if(F[j]>p) p=F[j];                   /* deletion  */
      
      S[j] = p;
      
    }
    
    /* transform score */
    transf_score = S[m] + minAS;

    scores->vals[i-1] = transf_score;      /* end position on db with score */
    scores->start[i-1]  = i - 1 - m - G[m] + 1;    /* start position with score */
    /* scores->vals[i-1] = S[m]; */       /* position on db with score, starts with 0! not 1 */
    score_distr[transf_score]++; /* sum up to transformed score distribution */

    /* rotate memory */
    tmp = S_; S_ = S; S = tmp;
    tmp = D_; D_ = D; D = tmp;
    tmp = F_; F_ = F; F = tmp;
    tmp = G_; G_ = G; G = tmp;
  }

  if(cfg.quiet == 0) {
    fprintf(stderr, "\r#    %-5.2f %% (db length: %d) %c strand",  100., n, strand);
    fflush(stderr);
  }
  free(S);free(S_);free(D);free(D_);free(F);free(F_); free(G); free(G_);
  
  return;
  
}

/* -------------------------------------------------------------------------- */

float * gotoh_bwRecursion2(char * s, char * q, int m, int s_pos, gotoh_settings cfg,
			   char *ss, char *qq, int sollscore, int n2add, int sA, int eA) {

  int i,j,k,len, p, pp;
  int **S, **D, **F;
  char *sa, *qa, /* *sx, */ btstate;
  char *db_part=NULL;
  int si,qj,found, start,gos,mms,perc,gquery;
  float *res=NULL;
  int db_end;

  if(sA < 0) sA = 0;
  
  len = eA - sA + 1;
  db_part = (char *)strndup(s+sA, len); /* printf("db: %s\n", db_part); */
  /* printf("qu: %s\n", q); */
  
  int minAS = abs( 2 * (cfg.gap_open + ((m - 1) * cfg.gap_extend)));

  /* printf("sollscore: %d  ",sollscore); */
  sollscore -= minAS;
  /* printf("retransformed: %d\n", sollscore); */
  
  S = (int **) calloc(len+2, sizeof(int *));
  D = (int **) calloc(len+2, sizeof(int *));
  F = (int **) calloc(len+2, sizeof(int *));
  for(i=0;i<=len;i++) {
    S[i] = (int *) calloc(m+2, sizeof(int *));
    D[i] = (int *) calloc(m+2, sizeof(int *));
    F[i] = (int *) calloc(m+2, sizeof(int *));
  }
  sa = (char *) calloc(m+len+2, sizeof(int *));
  qa = (char *) calloc(m+len+2, sizeof(int *));

  D[0][0] = (-cfg.inf);
  F[0][0] = (-cfg.inf);
  S[0][0] = 0;

  for(j=1;j<=m;j++) { 
    D[0][j] = (-cfg.inf);
    F[0][j] = (j-1)*cfg.gap_extend + cfg.gap_open;
    S[0][j] = F[0][j];
  }

  for(i=1;i<=len;i++) {
    si = sA + i - 1;

    D[i][0] = 0;
    F[i][0] = (-cfg.inf);
    S[i][0] = 0;
    
    for(j=1;j<=m;j++) {
      qj = j-1;	 

      p  = S[i][j-1] + cfg.gap_open;
      pp = D[i][j-1] + cfg.gap_extend;
      D[i][j] = p > pp ? p : pp;
      
      p  = S[i-1][j] + cfg.gap_open;
      pp = F[i-1][j] + cfg.gap_extend;
      F[i][j] = p > pp ? p : pp;

      p = gotoh_Table[(unsigned char)s[si]][(unsigned char)q[qj]];
      /* p = gotoh_Table[(unsigned char)db_part[i-1]][(unsigned char)q[qj]]; */

      p += S[i-1][j-1];

      if(D[i][j]>p) p=D[i][j];
      if(F[i][j]>p) p=F[i][j];
      
      S[i][j] = p;
    }
			 
  }

  /* printf("S[%d][%d]=%d\n",len,m,S[len][m]); */
  /* /\* check matrix *\/ */
  /* printf("\n"); */
  /* for(i=0; i<=len; i++) { */
  /*   printf("%3d:\t",i); */
  /*    for(j=0; j<=m; j++) { */
  /*     /\* printf("  %5.2f ",S[i][j]); *\/ */
  /*      printf("[%3d:%3d]={%3d,%3d,%3d}(%c%c)\t",i,j,S[i][j],D[i][j],F[i][j],s[sA+i-1],q[j]); */
      
  /*   } */
  /*    printf("\n"); */
  /* } */
  /* fflush(stdout); */
  /* ^^^ */

  int maxS = S[0][m];
  int maxI = 0;
  i = len;
  while(i>=1) {
    if(S[i][m] > maxS) {
      maxI = i; maxS = S[i][m];
    }
    i--;
  }

  /* printf("maxI: %d maxS: %d\n\n",maxI,maxS); */
 
  /* backtracing */
  res = (float *) calloc(7, sizeof(float));
  /* percId, mm, go, length */
  gos = 0;
  mms = 0;
  perc = 0;
  gquery = 0;

  i = maxI;
  j = m;
  k = 0;
  btstate = 'S';

  /* fprintf(stderr,"S[%d][%d] = %d (%d)\n",i,j,S[i][j],S[len][m]); /\* this is still correct *\/ */

  /* fprintf(stderr, "%s\n%s\n",db_part,q); */
  
  while( (j>1) ) {

    if(i<1) break;
    
    si = i-1; /*len -(i-1);*/
    qj = j-1; /* m -j; */

    /* if(si < 0) { */
    /*   while(j>0) { */
    /* 	sa[k] = '-'; */
    /* 	qa[k] = q[qj]; */
    /* 	start++; */
    /* 	j--; */
    /* 	k++; */
    /* 	qj = m -j; */
    /* 	si = 1; */
    /*   } */
    /*   break; */
    /* } */

    switch(btstate) {
    case 'S':
      /* fprintf(stderr, "%d:%c %d:%c = %d\n",i,db_part[si],j,q[qj],gotoh_Table[(unsigned char)db_part[si]][(unsigned char)q[qj]]); */
      if(S[i][j] == S[i-1][j-1] + gotoh_Table[(unsigned char)db_part[si]][(unsigned char)q[qj]]) {
	sa[k] = db_part[si];
	qa[k] = q[qj];
	db_end++;
	i--;
	j--;
	k++;
	btstate = 'S';
	if(gotoh_Table[(unsigned char)db_part[si]][(unsigned char)q[qj]] <= gotoh_Table[78][78])
	  mms++;
	else
	  perc++;
      }
      else if( S[i][j] == F[i][j] ) {
	btstate = 'F';
      }
      else if( S[i][j] == D[i][j] ) {
	btstate = 'D';
      }
      else {
	fprintf(stderr,"This should not happen in S[%d][%d]=%d (single seq bw 1, istscore=%d)\n",i,j,S[i][j],maxS);
	fprintf(stderr,"S[%d][%d]=%d\tF=%d\tD=%d\n",i,j,S[i][j],F[i][j],D[i][j]);
	fprintf(stderr,"S[%d][%d]=%d\tF=%d\tD=%d\n",i-1,j-1,S[i-1][j-1],F[i-1][j-1],D[i-1][j-1]);
	fprintf(stderr,"S[%d][%d]=%d\tF=%d\tD=%d\n",i,j-1,S[i][j-1],F[i][j-1],D[i][j-1]);
	fprintf(stderr,"S[%d][%d]=%d\tF=%d\tD=%d\n",i-1,j,S[i-1][j],F[i-1][j],D[i-1][j]);
	exit(0);
      } 
      break;
    case 'D':
      if(D[i][j] ==  S[i][j-1] + cfg.gap_open ) {
	sa[k] = '-';
	qa[k] = q[qj];
	start++;
        j--;	
	k++;
	btstate = 'S';
	gos++;
      }
      else if(D[i][j] == D[i][j-1] + cfg.gap_extend ) {
	sa[k] = '-';
	qa[k] = q[qj];
	start++;
	j--;
	k++;
	btstate = 'D';
      }
      else {
	fprintf(stderr,"This should not happen in D[%d][%d]=%d\n",i,j,D[i][j]);
	exit(0);
      } 
      break;
    case 'F':
      if(F[i][j] ==  S[i-1][j]   + cfg.gap_open ) {
	sa[k] = db_part[si];
	qa[k] = '-';
	db_end++;
	i--;
	k++;
	btstate = 'S';
	gos++;
	gquery++;
      }
      else if ( F[i][j] == F[i-1][j]  + cfg.gap_extend ) {
	sa[k] = s[si];
	qa[k] = '-';
	db_end++;
	i--;
	k++;
	btstate = 'F';
	gquery++;
      }
      else {
	fprintf(stderr,"This should not happen in F[%d][%d]=%d\n",i,j,F[i][j]);
	exit(0);
      }    
      break;
    default:
      fprintf(stderr,"This should not happen at all. btstate: %c\n",btstate);
      exit(0);
    }
    
  }

  k = strlen(sa);

  res[0] = (float) perc / (float) m * 100;
  res[1] = (float) mms;
  res[2] = (float) gos;
  res[3] = (float) k;
  res[4] = s_pos - perc - mms -gquery + 2;
  res[5] = s_pos + 1;

  strcpy(ss, sa);
  strcpy(qq, qa);

  /* printf("%s\n%s\n",ss,qq); */

  for(i=0;i<=len;i++) {
    free(S[i]);
    free(D[i]);
    free(F[i]);
  }
  free(S);free(D);free(F);
  free(sa);free(qa);

  free(db_part);
  
  return res;
}

/* -------------------------------------------------------------------------- */

float * gotoh_bwRecursion(char * s, char * q, int m, int s_pos, gotoh_settings cfg,
			  char *ss, char *qq, int sollscore, int n2add) {

  int i,j,k,len, p, pp;
  int **S, **D, **F;
  char *sa, *qa, /* *sx, */ btstate;
  int si,qj,found, start,gos,mms,perc,gquery;
  float *res;
  int db_end;

  len = 2 * m + 1;
  if(len > s_pos) len = s_pos+1; /* arrays start with 1, seq with 0 ! */
  if(s_pos - len < n2add) len = s_pos - n2add;
  if(len < m) return NULL;
  int minAS = abs( 2 * (cfg.gap_open + ((m - 1) * cfg.gap_extend)));
  /* retransform score */

  S = (int **) calloc(len+2, sizeof(int *));
  D = (int **) calloc(len+2, sizeof(int *));
  F = (int **) calloc(len+2, sizeof(int *));
  for(i=0;i<=len;i++) {
    S[i] = (int *) calloc(m+2, sizeof(int *));
    D[i] = (int *) calloc(m+2, sizeof(int *));
    F[i] = (int *) calloc(m+2, sizeof(int *));
  }
  sa = (char *) calloc(2*m+len+2, sizeof(int *));
  qa = (char *) calloc(2*m+len+2, sizeof(int *));

  D[0][0] = (-cfg.inf);
  F[0][0] = (-cfg.inf);
  S[0][0] = 0;

  for(j=1;j<=m;j++) { 
    D[0][j] = (-cfg.inf);
    F[0][j] = (j-1)*cfg.gap_extend + cfg.gap_open;
    S[0][j] = F[0][j];
  }

  for(i=1;i<=len;i++) {
    si = s_pos-(i-1);

    D[i][0] = 0;
    F[i][0] = (-cfg.inf);
    S[i][0] = 0;
    
    for(j=1;j<=m;j++) {
      qj = m-j;	 

      p  = S[i][j-1] + cfg.gap_open;
      pp = D[i][j-1] + cfg.gap_extend;
      D[i][j] = p > pp ? p : pp;
      
      p  = S[i-1][j] + cfg.gap_open;
      pp = F[i-1][j] + cfg.gap_extend;
      F[i][j] = p > pp ? p : pp;

      p = gotoh_Table[(unsigned char)s[si]][(unsigned char)q[qj]];

      p += S[i-1][j-1];

      if(F[i][j]>p) p=F[i][j];
      if(D[i][j]>p) p=D[i][j];
      
      S[i][j] = p;
    }
			 
  }

  /* printf("S[%d][%d]=%d\n",len,m,S[len][m]); */
  /* /\* check matrix *\/ */
  /* printf("\n"); */
  /* for(i=0; i<=len; i++) { */
  /*   printf("%3d:\t",i); */
  /*    for(j=0; j<=m; j++) { */
  /*     /\* printf("  %5.2f ",S[i][j]); *\/ */
  /*     printf("[%3d:%3d]={%3d,%3d,%3d}\t",i,j,S[i][j],D[i][j],F[i][j]); */
  /*   } */
  /*    printf("\n"); */
  /* } */
  /* printf("sollscore: %d\n", sollscore); */
  /* fflush(stdout); */
  /* ^^^ */

  
  int maxscore = sollscore;
  int maxscoreI=0;
  
  i=len; found = 0;
  while(i>=1) {
    if(S[i][m] == sollscore) {
      /* maxscore = sollscore; */
      maxscoreI = i;
      found = 1;
      break;
    } else if(S[i][m] > maxscore) {
      maxscore = S[i][m];
      maxscoreI = i;
    }
    i--;
  }
  
  if(i == 0 && len < 2*m+1) { /* in this case, it overlaps with a better candidate */
    return NULL;
    fprintf(stderr, "sollscore %d not found in [%d:%d]\n",sollscore,s_pos-len,s_pos); exit(0);
  }
  
   /* i is now the position with our requested score !!! */
  i = maxscoreI;

  if(maxscore != sollscore) {
    fprintf(stderr, "NOTE: At position %d a better alignment then the requested one ",maxscoreI+n2add);
    fprintf(stderr, "at %d is possible.\n",s_pos);
    fprintf(stderr, "DS:  |%d-%d| = %d\n",maxscore,sollscore, abs(maxscore-sollscore));
    fprintf(stderr, "DSntf|%d-%d| = %d\n",maxscore+minAS,sollscore+minAS, abs(maxscore+minAS-(sollscore+minAS)));
    fprintf(stderr, "DDB: |%d-%d| = %d\n",maxscoreI+n2add,s_pos, abs(s_pos-(maxscoreI+n2add)));
    /* exit(0); */
  }

  int realend = i;
 
  start = s_pos-i+1;
  db_end = start;
  
  res = (float *) calloc(7, sizeof(float));
  /* percId, mm, go, length */
  gos = 0;
  mms = 0;
  perc = 0;
  gquery = 0;
  
  j = m;
  k = 0;
  btstate = 'S';


  while( (j>0) ) {

    if(i<1) break;
    
    si = s_pos -(i-1);
    qj = m -j;

    if(si < 0) {
      while(j>0) {
	sa[k] = '-';
	qa[k] = q[qj];
	start++;
	j--;
	k++;
	qj = m -j;
	si = 1;
      }
      break;
    }

    switch(btstate) {
    case 'S':
      if(S[i][j] == S[i-1][j-1] + gotoh_Table[(unsigned char)s[si]][(unsigned char)q[qj]]) {
	sa[k] = s[si];
	qa[k] = q[qj];
	db_end++;
	i--;
	j--;
	k++;
	btstate = 'S';
	if(gotoh_Table[(unsigned char)s[si]][(unsigned char)q[qj]] <= gotoh_Table[78][78])
	  mms++;
	else
	  perc++;
      }
      else if( S[i][j] == F[i][j] ) {
	btstate = 'F';
      }
      else if( S[i][j] == D[i][j] ) {
	btstate = 'D';
      }
      else {
	fprintf(stderr,"This should not happen in S[%d][%d]=%d (single seq bw 1, istscore=%d)\n",i,j,S[i][j],sollscore);
	exit(0);
      } 
      break;
    case 'D':
      if(D[i][j] ==  S[i][j-1] + cfg.gap_open ) {
	sa[k] = '-';
	qa[k] = q[qj];
	start++;
        j--;	
	k++;
	btstate = 'S';
	gos++;
      }
      else if(D[i][j] == D[i][j-1] + cfg.gap_extend ) {
	sa[k] = '-';
	qa[k] = q[qj];
	start++;
	j--;
	k++;
	btstate = 'D';
      }
      else {
	fprintf(stderr,"This should not happen in D[%d][%d]=%d\n",i,j,D[i][j]);
	exit(0);
      } 
      break;
    case 'F':
      if(F[i][j] ==  S[i-1][j]   + cfg.gap_open ) {
	sa[k] = s[si];
	qa[k] = '-';
	db_end++;
	i--;
	k++;
	btstate = 'S';
	gos++;
	gquery++;
      }
      else if ( F[i][j] == F[i-1][j]  + cfg.gap_extend ) {
	sa[k] = s[si];
	qa[k] = '-';
	db_end++;
	i--;
	k++;
	btstate = 'F';
	gquery++;
      }
      else {
	fprintf(stderr,"This should not happen in F[%d][%d]=%d\n",i,j,F[i][j]);
	exit(0);
      }    
      break;
    default:
      fprintf(stderr,"This should not happen at all. btstate: %c\n",btstate);
      exit(0);
    }
    
  }

  k = strlen(sa);

  res[0] = (float) perc / (float) m * 100;
  res[1] = (float) mms;
  res[2] = (float) gos;
  res[3] = (float) k;
  res[4] = s_pos - perc - mms -gquery + 2;
  res[5] = s_pos + 1;

  strcpy(ss, sa);
  strcpy(qq, qa);

  for(i=0;i<=len;i++) {
    free(S[i]);
    free(D[i]);
    free(F[i]);
  }
  free(S);free(D);free(F);
  free(sa);free(qa);

  return res;
  
}

/* -------------------------------------------------------------------------- */

void gotoh_findLocalMaxima(struct iarray *scores, gotoh_aln List[], int s, int e,
			   char *db_seq, char *db_name, char *q_seq, char *q_name, int m,
			   char strand, struct gotoh_settings gs, struct iarray *lm, int db_len) {
    
  int ms, msT, mi, nsl, nel, nsr, ner, ii, newend;

  if(s < 0 || e >= db_len) return;
  if(e - s < m) return;
  
  
  ms = 0; mi = -1;
  gotoh_getMaxLinear(scores, s, e, &ms, &mi); /* insert/backtrace ms an mi */
  lm->vals[ms] += 1;

  /* printf("M %d %d\n",ms,mi); */

  /* fprintf(stderr,"SE von %d bis %d maxScore %d at pos %d (%c)\n",s,e,ms, mi, strand); */

  ii=0; /* find place in L to put local maxima */
  
  while((int)(List[ii].alnscore) > ms && ii < MAX_ALNS) ii++;

  if(ii == MAX_ALNS) return; /* list already full with scores all > ms ! */

  /* try to prevent including duplicated entries !!*/
  /* newend = gotoh_backtraceAln2(ii, mi, ms, strand, db_name, q_name, List, m, db_seq, q_seq, gs, s, scores->size); */
  newend = gotoh_backtraceAln2(ii, mi, ms, strand, db_name, q_name, List, m, db_seq, q_seq, gs, s, scores);

  if(newend == -1) return;
  
  /* d&c recursion on left side */
  nsl = s;
  nel = newend - 1; 
  gotoh_findLocalMaxima(scores, List, nsl, nel, db_seq, db_name, q_seq, q_name, m, strand, gs, lm, db_len);
  
  /* d&c recursion on right side */
  nsr = mi + m + 1;
  ner = e;
  gotoh_findLocalMaxima(scores, List, nsr, ner, db_seq, db_name, q_seq, q_name, m, strand, gs, lm, db_len);
  
  return;

}

/* -------------------------------------------------------------------------- */
void gotoh_getMaxLinear(struct iarray *scores, int s, int e, int *ms, int *mi) {
  
  int i, j, k, sm=0, im=-1, prev;

  i = s;
  prev = scores->vals[i];
  while(scores->vals[i] <= prev && i<=e && i<scores->size) {
    prev = scores->vals[i];
    i++;
  }

  j = e;
  prev = scores->vals[j];
  while(scores->vals[j] <= prev && j>=s && j>=0) {
    prev = scores->vals[j];
    j--;
  }

  if(i>=j) return;
  
  for(k=i; k<=j; k++) {
    if(scores->vals[k] >= sm) {
      sm = scores->vals[k];
      im = k;
    }
  }

  *ms = sm; *mi = im;
  
  return;
  
}

/* -------------------------------------------------------------------------- */
/**
   q_seq is sense if strand eq '+', otherwise it is already in antisense 
 */
int gotoh_backtraceAln2(int ii, int imax, int smax, char strand, char *s_name, char* q_name,
                         gotoh_aln List[], int m, char *db_seq, char *q_seq,
                         gotoh_settings gs, int n2add, struct iarray *scores) {

  float *bt;
  gotoh_aln curaln;
  int g,shift, start=0,Len, sA, eA;

  Len = scores->size;
  
  sA = scores->start[imax];
  if(sA < 0) sA = 0;
  eA = imax;
  
  shift = abs(2 * (gs.gap_open + (m-1) * gs.gap_extend));
  
  curaln = gotoh_initL();

  curaln.alnscore = (float)(smax);
  curaln.ntfalnscore = (float) (smax - shift);
  curaln.db_len = Len;
  curaln.qStart = 1;
  curaln.qEnd = m;
  curaln.strand = strand;
  curaln.hitname = (char *) calloc(strlen(s_name)+1, sizeof(char));
  strcpy(curaln.hitname, s_name);
  curaln.queryname = (char *) calloc(strlen(q_name)+1, sizeof(char));
  strcpy(curaln.queryname, q_name);
  curaln.hitseq = (char *) calloc(4*m+1, sizeof(char));
  curaln.queryseq = (char *) calloc(2*m+1, sizeof(char));

  bt = gotoh_bwRecursion2(db_seq, q_seq, m, imax, gs, curaln.hitseq, curaln.queryseq, smax, n2add, sA, eA);

  if(bt == NULL) return -1;
  
  curaln.percID = bt[0]; curaln.mismatches = (int) bt[1];
  curaln.gapopening = (int) bt[2]; curaln.length = (int) bt[3];
  if(strand == '-') {
    curaln.sStart = Len - (int)(bt[5]);
    curaln.sEnd   = Len - (int)(bt[4]);
  } else {
    curaln.sStart = (int)(bt[4]);
    curaln.sEnd   = (int)(bt[5]);
  }
  curaln.evalue = m * Len * exp(smax);
  start = (int)(bt[4]);
  free(bt);

  /* don't insert duplicated entry (due to splits system!) */
  if(ii>=1 && List[ii-1].alnscore == curaln.alnscore && List[ii-1].sEnd == curaln.sEnd) {
    return start;
  }

  if(List[ii].alnscore == -1) { /* empty position in List */
    gotoh_copyGotohAln(&List[ii], curaln);
    return start;
  } 

  /* position not empty, move all coming entries one pos backward */
  for(g=MAX_ALNS-1; g>ii; g--) {
    if(List[g-1].alnscore == -1) continue;

    List[g] = List[g-1];
    
  }
  gotoh_copyGotohAln(&List[ii], curaln);

  gotoh_freeGotohAln(curaln);

  return start;
  
}

/* -------------------------------------------------------------------------- */

/* End of file */
